from typing import Type, Dict, Any, Optional, Tuple

import torch

from athlete.update.update_rule import UpdateRule, UpdatableComponent
from athlete.data_collection.provider import UpdateDataProvider
from athlete.algorithms.ppo.module import PPOActor
from athlete.saving.saveable_component import CompositeSaveableComponent
from athlete.algorithms.ppo.updatable_components import (
    PPOBufferUpdate,
    PPOGradientUpdate,
)
from athlete.update.on_policy_buffer import OnPolicyBuffer


class PPOUpdate(UpdateRule, CompositeSaveableComponent):
    """The Update rule for PPO.
    This class manages all updatable components and the saving and loading of stateful objects.
    """

    def __init__(
        self,
        update_data_provider: UpdateDataProvider,
        value_function: torch.nn.Module,
        actor: PPOActor,
        optimizer_class: Type[torch.optim.Optimizer],
        optimizer_arguments: Dict[str, Any],
        discount: float,
        generalized_advantage_estimation_lambda: float,
        epochs_per_update: int,
        mini_batch_size: int,
        policy_ratio_clip_value: float,
        value_loss_clip_value: Optional[float],
        value_loss_coefficient: float,
        entropy_loss_coefficient: float,
        batch_normalize_advantage: bool,
        gradient_max_norm: Optional[float],
        device: str,
    ):
        """Initializes the PPO update rule.

        Args:
            update_data_provider (UpdateDataProvider): The data provider used to communicate with the data collector.
            value_function (torch.nn.Module): The value function network.
            actor (PPOActor): The actor network.
            optimizer_class (Type[torch.optim.Optimizer]): The optimizer class to use for the value function and actor networks.
            optimizer_arguments (Dict[str, Any]): The initialization arguments for the optimizer of actor and value function. Network parameters are added automatically.
            discount (float): The discount factor used for return calculation.
            generalized_advantage_estimation_lambda (float): The lambda parameter for generalized advantage estimation.
            epochs_per_update (int): The number of epochs to perform for each update.
            mini_batch_size (int): The mini-batch size for sampling from the buffer.
            policy_ratio_clip_value (float): The clip value for the policy ratio.
            value_loss_clip_value (Optional[float]): The clip value for the value loss. None means no clipping.
            value_loss_coefficient (float): The coefficient for the value loss in the total loss.
            entropy_loss_coefficient (float): The coefficient for the entropy loss in the total loss.
            batch_normalize_advantage (bool): Whether to batch normalize the advantage before calculating the loss.
            gradient_max_norm (Optional[float]): The maximum norm for gradient clipping. None means no clipping.
            device (str): The device to use for the networks (e.g., "cpu" or "cuda").
        """
        UpdateRule.__init__(self)
        CompositeSaveableComponent.__init__(self)

        self.value_function = value_function.to(device)
        self.actor = actor.to(device)

        self.optimizer = optimizer_class(
            [
                {
                    "params": self.actor.parameters(),
                    **optimizer_arguments,
                },
                {
                    "params": self.value_function.parameters(),
                    **optimizer_arguments,
                },
            ]
        )

        self.register_saveable_component("value_function", self.value_function)
        self.register_saveable_component("actor", self.actor)
        self.register_saveable_component("optimizer", self.optimizer)

        # PPO Buffer Update
        self.buffer = OnPolicyBuffer(mini_batch_size=mini_batch_size)

        self.buffer_update = PPOBufferUpdate(
            update_data_provider=update_data_provider,
            value_function=self.value_function,
            discount=discount,
            generalized_advantage_estimation_lambda=generalized_advantage_estimation_lambda,
            on_policy_buffer=self.buffer,
            changes_policy=False,
            save_file_name=PPOBufferUpdate.SAVE_FILE_NAME,
        )

        self.register_saveable_component("ppo_buffer_update", self.buffer_update)

        # PPO Gradient Update

        self.gradient_update = PPOGradientUpdate(
            value_function=self.value_function,
            actor=self.actor,
            optimizer=self.optimizer,
            on_policy_buffer=self.buffer,
            epochs_per_update=epochs_per_update,
            policy_ratio_clip_value=policy_ratio_clip_value,
            value_loss_clip_value=value_loss_clip_value,
            value_loss_coefficient=value_loss_coefficient,
            entropy_loss_coefficient=entropy_loss_coefficient,
            batch_normalize_advantage=batch_normalize_advantage,
            gradient_max_norm=gradient_max_norm,
            policy_loss_log_tag=PPOGradientUpdate.POLICY_LOSS_LOG_TAG,
            value_loss_log_tag=PPOGradientUpdate.VALUE_LOSS_LOG_TAG,
            entropy_loss_log_tag=PPOGradientUpdate.ENTROPY_LOSS_LOG_TAG,
            total_loss_log_tag=PPOGradientUpdate.TOTAL_LOSS_LOG_TAG,
            changes_policy=True,
            save_file_name=PPOGradientUpdate.SAVE_FILE_NAME,
        )

        self.register_saveable_component("ppo_gradient_update", self.gradient_update)

    @property
    def updatable_components(self) -> Tuple[UpdatableComponent]:
        """Returns a tuple of all updatable components of the PPO update rule in the order they should be updated in.

        Returns:
            Tuple[UpdatableComponent]: The updatable components of the PPO update rule:
                1. PPO buffer update
                2. PPO gradient update
        """
        return self.buffer_update, self.gradient_update



from typing import Tuple, Any
from abc import ABC, abstractmethod
import os

import numpy as np

from athlete.global_objects import RNGHandler
from athlete.saving.saveable_component import SaveContext


class NoiseProcess(ABC):
    """Interface for noise processes."""

    def __init__(self) -> None:
        super().__init__()

    def reset(self) -> None:
        """Overwrite this if the noise process has any state that needs to be reset upon resetting the environment."""
        pass

    @abstractmethod
    def sample(self) -> Any:
        """Sample noise from the noise process.

        Returns:
            Any: The sampled noise. The type of the noise depends on the implementation of the noise process.
        """
        ...

    def save_checkpoint(self, context: SaveContext) -> None:
        """Overwrite this if the noise process has any state that needs to be saved.

        Args:
            context (SaveContext): The context object that contains the save path and file handler.
        """
        pass

    def load_checkpoint(self, context: SaveContext) -> None:
        """Overwrite this if the noise process has any state that needs to be saved.

        Args:
            context (SaveContext): The context object that contains the save path and file handler.
        """
        pass


class GaussianNoise(NoiseProcess):
    """A Gaussian noise process."""

    def __init__(
        self,
        shape: Tuple[int, ...],
        noise_std: float = 0.2,
    ) -> None:
        """Initialize the Gaussian noise process.

        Args:
            shape (Tuple[int, ...]): The shape of the noise to be sampled.
            noise_std (float, optional): The standard deviation of the Gaussian noise. Defaults to 0.2.
        """
        super().__init__()
        self.shape = shape
        self.noise_std = noise_std
        self.random_numbers_generator = RNGHandler.get_random_number_generator()

    def sample(self) -> np.ndarray:
        """Sample noise from the Gaussian noise process.

        Returns:
            np.ndarray: The sampled Gaussian noise.
        """
        return self.random_numbers_generator.normal(
            loc=0.0, scale=self.noise_std, size=self.shape
        )


class OhrsteinUhlenbeckNoise(NoiseProcess):
    """A class that implements the Ornstein-Uhlenbeck process for generating noise."""

    SAVE_FILE_NAME = "ohrnstein_uhlenbeck_noise"

    def __init__(
        self,
        shape: Tuple[int, ...],
        noise_std: float = 0.2,
        theta: float = 0.15,
        dt: float = 0.01,
        mu: float = 0.0,
    ) -> None:
        """Initialize the Ornstein-Uhlenbeck noise process.

        Args:
            shape (Tuple[int, ...]): The shape of the noise to be sampled.
            noise_std (float, optional): The standard deviation of the Gaussian noise. Defaults to 0.2.
            theta (float, optional): The rate of mean reversion. Defaults to 0.15.
            dt (float, optional): The time increment per sample, the lower this is the smoother the process becomes. Defaults to 0.01.
            mu (float, optional): The long-term mean of the process. Defaults to 0.0.
        """
        super().__init__()
        self.noise_std = noise_std
        self.theta = theta
        self.dt = dt
        self.mu = mu
        self.shape = shape
        self.last_noise = np.zeros(shape)
        self.random_numbers_generator = RNGHandler.get_random_number_generator()

    def reset(self) -> None:
        """Reset the noise process to its initial state which means it samples near the mean of the process."""
        self.last_noise = np.zeros(self.shape)

    def sample(self) -> np.ndarray:
        """Sample noise from the Ornstein-Uhlenbeck process.

        Returns:
            np.ndarray: The sampled noise from the Ornstein-Uhlenbeck process.
        """
        noise = (
            self.last_noise
            + self.theta * (self.mu - self.last_noise) * self.dt
            + self.noise_std
            * np.sqrt(self.dt)
            * self.random_numbers_generator.normal(size=self.shape)
        )
        self.last_noise = noise
        return noise

    def save_checkpoint(self, context: SaveContext) -> None:
        """Saving the last noise state of the Ornstein-Uhlenbeck process."""
        full_save_path = os.path.join(
            context.save_path, context.prefix + self.SAVE_FILE_NAME
        )

        handling_stats = (self.last_noise,)

        context.file_handler.save_to_file(
            to_save=handling_stats, save_path=full_save_path
        )

    def load_checkpoint(self, context: SaveContext) -> None:
        """Load the last noise state of the Ornstein-Uhlenbeck process."""
        full_load_path = os.path.join(
            context.save_path, context.prefix + self.SAVE_FILE_NAME
        )

        handling_stats = context.file_handler.load_from_file(load_path=full_load_path)

        self.last_noise = handling_stats[0]


class ColoredNoise(NoiseProcess):
    """A class that implements a colored noise process. A noise type that filters high frequencies out such that the noise is smoother than white noise."""

    SAVE_FILE_NAME = "colored_noise"

    def __init__(
        self,
        shape: Tuple[int, ...],
        horizon: int,
        beta: float = 1.0,
        noise_std: float = 1.0,
    ) -> None:
        """Initialize the colored noise process.

        Args:
            shape (Tuple[int, ...]): The shape of the noise to be sampled.
            horizon (int): The length of the generated noise sequence. If the sequence is depleted, a new one is generated.
            beta (float, optional): The exponent for the frequency scaling. A value of 0 corresponds to white noise, while a value of 2 corresponds to Brownian noise. Defaults to 1.0.
            noise_std (float, optional): The standard deviation of the Gaussian noise. Defaults to 1.0.
        """
        super().__init__()
        self.shape = shape
        self.horizon = horizon
        self.beta = beta
        self.sequence_buffer = np.zeros((horizon, *shape))
        self.pointer = 0
        self.noise_std = noise_std
        self.random_numbers_generator = RNGHandler.get_random_number_generator()

    def reset(self) -> None:
        """This will cause a new sequence to be generated on the next sample call"""
        self.pointer = 0

    def sample(self) -> np.ndarray:
        """Sample noise from the colored noise process.
        The noise is sampled from a pre-generated sequence. If the sequence is depleted, a new one is generated.

        Returns:
            np.ndarray: The sampled noise from the colored noise process.
        """
        if self.pointer == 0:
            self.sequence_buffer = self._generate_colored_noise_sequence(
                beta=self.beta,
                sequence_length=self.horizon,
                dimensions=np.prod(self.shape),
            ).T  # transposing so that first dimension is the time dimension second are the dimensions of the action space

        noise = self.noise_std * self.sequence_buffer[self.pointer]
        self.pointer = (self.pointer + 1) % self.horizon

        return noise

    def _generate_colored_noise_sequence(
        self, beta, sequence_length, dimensions=1
    ) -> np.ndarray:
        """Generate a sequence of colored noise using the Fourier method.
        The noise is generated in the frequency domain and then transformed to the time domain using the inverse FFT.
        The generated noise is normalized to have a standard deviation of 1.

        Args:
            beta (_type_): The exponent for the frequency scaling. A value of 0 corresponds to white noise, while a value of 2 corresponds to Brownian noise.
            sequence_length (_type_): The length of the generated noise sequence.
            dimensions (int, optional): The number of dimensions of the generated noise. Defaults to 1.

        Returns:
            np.ndarray: The generated colored noise sequence.
        """
        # Create frequency bins
        frequency_amplitudes = np.fft.rfftfreq(sequence_length)

        num_positive_frequencies = frequency_amplitudes.shape[0]

        # Create the scaling factors for each frequency
        frequencies_scaling = np.zeros(num_positive_frequencies)
        frequencies_scaling[1:] = np.power(
            np.abs(frequency_amplitudes[1:]), -beta / 2
        )  # Avoid division by zero

        # Calculate the theoretical standard deviation for scaling
        sigma = np.sqrt(np.sum(frequencies_scaling**2) / num_positive_frequencies)

        # Generate random phases. Introduce a random phase shift for each dimension.
        phases = self.random_numbers_generator.uniform(
            low=0, high=2 * np.pi, size=(dimensions, num_positive_frequencies)
        )

        # Create the complex frequency components. We generate them across all dimensions
        signal_frequencies = frequencies_scaling * np.exp(1j * phases)

        # Inverse FFT to get the time-domain signal, for each dimension
        signal = np.fft.irfft(signal_frequencies, n=sequence_length)

        # Normalize the signal
        signal /= sigma

        return signal

    def save_checkpoint(self, context: SaveContext) -> None:
        """Saves the current sequence buffer and pointer of the colored noise process.

        Args:
            context (SaveContext): The context object that contains the save path and file handler.
        """
        full_save_path = os.path.join(
            context.save_path, context.prefix + self.SAVE_FILE_NAME
        )

        handling_stats = (
            self.sequence_buffer,
            self.pointer,
        )

        context.file_handler.save_to_file(
            to_save=handling_stats, save_path=full_save_path
        )

    def load_checkpoint(self, context: SaveContext) -> None:
        """Loads the current sequence buffer and pointer of the colored noise process.

        Args:
            context (SaveContext): The context object that contains the save path and file handler.
        """
        full_load_path = os.path.join(
            context.save_path, context.prefix + self.SAVE_FILE_NAME
        )

        handling_stats = context.file_handler.load_from_file(load_path=full_load_path)

        (
            self.sequence_buffer,
            self.pointer,
        ) = handling_stats



from typing import Any, Dict, Tuple
import numpy as np
import torch
from gymnasium.spaces import Box

from athlete.function import numpy_to_tensor, tensor_to_numpy
from athlete.policy.policy_builder import Policy, PolicyBuilder
from athlete.algorithms.ppo.module import PPOActor

INFO_KEY_LOG_PROB = "log_prob"


class PPOTrainingPolicy(Policy):
    """Training policy implementation for PPO algorithm.

    This policy uses a PPO actor network to sample actions during training and record log probabilities of taken actions.
    """

    def __init__(
        self,
        actor: PPOActor,
        action_space: Box,
    ) -> None:
        """Initializes the PPO training policy.

        Args:
            actor (PPOActor): The PPO actor to use for action selection.
            action_space (Box): The action space of the actor.
        """
        self.actor = actor
        self.action_space = action_space

        self.action_high = action_space.high
        self.action_low = action_space.low

        self.module_device = next(self.actor.parameters()).device

    def act(self, observation: np.ndarray) -> Tuple[int, Dict[str, Any]]:
        """Samples an action from the actor given the observation.

        Args:
            observation (np.ndarray): The observation from the environment.

        Returns:
            Tuple[int, Dict[str, Any]]: A tuple containing the sampled action and a dictionary containing the log probability of the action.
        """
        observation = np.expand_dims(observation, axis=0)
        observation = numpy_to_tensor(observation, device=self.module_device)
        with torch.no_grad():
            action, log_prob = self.actor.get_action_and_log_prob(observation)
        action = tensor_to_numpy(action).squeeze(axis=0)
        log_prob = tensor_to_numpy(log_prob).squeeze(axis=0)

        return action, {INFO_KEY_LOG_PROB: log_prob}


class PPOEvaluationPolicy(Policy):
    """The PPO evaluation policy.
    This policy uses the actor to get the deterministic action mean.
    """

    def __init__(
        self,
        actor: PPOActor,
        action_space: Box,
    ) -> None:
        """Initializes the PPO evaluation policy.

        Args:
            actor (PPOActor): The PPO actor to use for action selection.
            action_space (Box): The action space of the actor.
        """
        self.actor = actor
        self.action_space = action_space

        self.action_high = action_space.high
        self.action_low = action_space.low

        self.module_device = next(self.actor.parameters()).device

    def act(self, observation: np.ndarray) -> Tuple[int, Dict[str, Any]]:
        """Get the deterministic action mean from the actor given the observation.

        Args:
            observation (np.ndarray): The observation from the environment.

        Returns:
            Tuple[int, Dict[str, Any]]: A tuple containing the deterministic action mean and an empty dictionary.
        """
        observation = np.expand_dims(observation, axis=0)
        observation = numpy_to_tensor(observation, device=self.module_device)
        with torch.no_grad():
            action = self.actor.get_mean(observation)
        action = tensor_to_numpy(action).squeeze(axis=0)

        return action, {}


class PPOPolicyBuilder(PolicyBuilder):
    """Policy builder for PPO.
    This class is responsible for creating the training and evaluation policies for PPO.
    """

    def __init__(
        self,
        actor: PPOActor,
        action_space: Box,
    ) -> None:
        """Initializes the PPO policy builder.

        Args:
            actor (PPOActor): The PPO actor to use for action selection.
            action_space (Box): The action space of the actor.
        """
        super().__init__()
        self.actor = actor
        self.action_space = action_space

    def build_training_policy(self) -> Policy:
        """Builds the training policy for PPO.

        Returns:
            Policy: The PPO training policy.
        """
        return PPOTrainingPolicy(
            actor=self.actor,
            action_space=self.action_space,
        )

    def build_evaluation_policy(self) -> Policy:
        """Builds the evaluation policy for PPO.

        Returns:
            Policy: The PPO evaluation policy.
        """
        return PPOEvaluationPolicy(
            actor=self.actor,
            action_space=self.action_space,
        )

    @property
    def requires_rebuild_on_policy_change(self) -> bool:
        """Whether the policy requires a rebuild on policy change.

        Returns:
            bool: False, as the update only changes the weights of the actor which are
            referenced by the policy classes.
        """
        return False



from typing import Dict, Any, List, Optional
import numpy as np
from operator import itemgetter

from athlete.update.buffer import Buffer
from athlete.saving.file_handler import FileHandler
from athlete.function import chain_functions
from athlete.saving.saveable_component import SaveContext


class BufferWrapper(Buffer):
    """Base class for buffer wrappers. Useful if data should be manipulated in some way when being and or sampled from the buffer.
    Simply mimics the behavior of the wrapped buffer if no function is overwritten."""

    def __init__(self, replay_buffer: Buffer) -> None:
        """Initializes the BufferWrapper with a replay buffer.

        Args:
            replay_buffer (Buffer): The replay buffer to be wrapped.
        """
        self.replay_buffer = replay_buffer

    def add(
        self, transition_dictionary: Dict[str, np.ndarray], episode_ended: np.ndarray
    ) -> None:
        self.replay_buffer.add(
            data_dictionary=transition_dictionary, episode_ended=episode_ended
        )

    def sample(self, batch_size: int) -> Dict[str, np.ndarray]:
        return self.replay_buffer.sample(batch_size=batch_size)

    def encode_sample(self, sample_ids: List[int]) -> Dict[str, np.ndarray]:
        return self.replay_buffer.encode_sample(sample_ids)

    def save_checkpoint(self, context: SaveContext):
        self.replay_buffer.save_checkpoint(
            context=context,
        )

    def load_checkpoint(self, file_handler: FileHandler):
        self.replay_buffer.load_checkpoint(
            file_handler=file_handler,
        )

    @property
    def size(self):
        return self.replay_buffer.size

    @property
    def max_size(self):
        return self.replay_buffer.max_size

    @property
    def data_info(self) -> Dict[str, Dict[str, Any]]:
        return self.replay_buffer.data_info

    @property
    def pointer_position(self):
        return self.replay_buffer.pointer_position


class InputOutputWrapper(BufferWrapper):
    """Baseclass for specific wrapper type that does some modification during adding and sampling.
    This is useful for things like normalization, where you want to update statistics on adding,
    and then normalize data upon sampling.
    """

    def __init__(
        self,
        replay_buffer: Buffer,
        data_info: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> None:
        super().__init__(replay_buffer=replay_buffer)

        if data_info is None:
            self._data_info = replay_buffer.data_info
        else:
            self._data_info = data_info

    def add(
        self, transition_dictionary: Dict[str, np.ndarray], episode_ended: np.ndarray
    ) -> None:
        transformed = self.in_transform(data_dictionary=transition_dictionary)
        self.replay_buffer.add(data_dictionary=transformed, episode_ended=episode_ended)
        self.post_add_routine()

    def sample(self, batch_size: int) -> Dict[str, np.ndarray]:
        raw_sample = self.replay_buffer.sample(batch_size=batch_size)
        transformed = self.out_transform(data_dictionary=raw_sample)
        return transformed

    def in_transform(self, data_dictionary: Dict[str, Any]) -> Dict[str, Any]:
        """Overwrite this function to transform the data before adding it to the buffer.

        Args:
            data_dictionary (Dict[str, Any]): The data to be transformed.

        Returns:
            Dict[str, Any]: The transformed data.
        """
        return data_dictionary

    def post_add_routine(self):
        """This function is called after adding data to the buffer."""
        pass

    def out_transform(self, data_dictionary: Dict[str, Any]) -> Dict[str, Any]:
        """Overwrite this function to transform the data after sampling it from the buffer.

        Args:
            data_dictionary (Dict[str, Any]): The sampled data to be transformed.

        Returns:
            Dict[str, Any]: Transformed sampled data.
        """
        return data_dictionary


class PostBufferPreprocessingWrapper(InputOutputWrapper):
    """A wrapper that applies preprocessing functions to the data after sampling from the buffer.
    This is useful if the data is stored in a different format than it is used for training.
    E.g. using int8 values to save memory of pixel inputs and then converting them to float32 for training.
    """

    def __init__(
        self,
        replay_buffer: Buffer,
        post_replay_buffer_data_preprocessing: Dict[str, List[callable]],
    ) -> None:
        """Initializes the PostBufferPreprocessingWrapper with a replay buffer and a dictionary of preprocessing functions.
        The keys of the dictionary are the names of the fields in the replay buffer, and the values are lists of functions to be applied to the corresponding field.

        Args:
            replay_buffer (Buffer): The replay buffer to be wrapped.
            post_replay_buffer_data_preprocessing (Dict[str, List[callable]]): Dictionary of preprocessing functions to be applied to the data after sampling from the buffer.
                The keys of the dictionary are the names of the fields in the replay buffer, and the values are lists of functions to be applied to the corresponding field.
        """
        super().__init__(replay_buffer=replay_buffer)
        self.post_replay_buffer_preprocessing = post_replay_buffer_data_preprocessing
        self.field_getter = itemgetter(
            *list(post_replay_buffer_data_preprocessing.keys())
        )

    def out_transform(self, data_dictionary: Dict[str, Any]) -> Dict[str, Any]:
        """Applies the preprocessing functions to the data after sampling from the buffer.

        Args:
            data_dictionary (Dict[str, Any]): The sampled data to be transformed.

        Returns:
            Dict[str, Any]: Transformed sampled data.
        """

        preprocessed_values = map(
            chain_functions,
            self.post_replay_buffer_preprocessing.values(),
            self.field_getter(data_dictionary),
        )

        preprocessed_dictionary = dict(
            zip(self.post_replay_buffer_preprocessing.keys(), preprocessed_values)
        )

        data_dictionary.update(preprocessed_dictionary)

        return data_dictionary



import os
from typing import Dict, Any, Generator

from athlete.global_objects import RNGHandler
from athlete.saving.saveable_component import SaveContext


class OnPolicyBuffer:
    """A simple buffer for storing arbitrary data needed by on-policy algorithms.

    This buffer stores experience batches for on-policy algorithms like PPO and provides
    functionality to generate shuffled mini-batches during policy updates. Unlike
    replay buffers used in off-policy algorithms, this buffer is cleared and refilled
    during each update cycle.
    """

    SAVE_FILE_NAME = "on_policy_buffer"
    FILE_EMPTY_SIGNAL = "empty"

    def __init__(
        self,
        mini_batch_size: int,
        save_file_name: str = SAVE_FILE_NAME,
    ) -> None:
        """Initializes the OnPolicyBuffer with a mini batch size and a save file name.

        Args:
            mini_batch_size (int): Size of the mini batches to be generated. If this is not a fraction of the total size, the last batch will be smaller.
            save_file_name (str, optional): Name of the file to save the buffer to. Defaults to "on_policy_buffer".
        """

        self.data = {}
        self.size = 0
        self.all_keys = []
        self.mini_batch_size = mini_batch_size
        self.random_number_generator = RNGHandler.get_random_number_generator()

        self.save_file_name = save_file_name

    def set_data(self, data_dict: Dict[str, Any]) -> None:
        """Set the data of the buffer. Old data will be overwritten.

        Args:
            data_dict (Dict[str, Any]): Dictionary containing the data to be stored in the buffer. The keys of the dictionary will be used to access the data.
        """
        self.data = data_dict
        self.all_keys = list(data_dict.keys())
        self.size = data_dict[self.all_keys[0]].shape[0]

    def generate_shuffled_batched_epoch(self) -> Generator[Dict[str, Any], None, None]:
        """Generates shuffled mini-batches of data from the buffer. One generator call will generate one epoch of data. The data is shuffled before being split into mini-batches.

        Yields:
            Generator[Dict[str, Any], None, None]: Generator yielding mini-batches of data. Each mini-batch is a dictionary containing the data for each key in the buffer.
        """

        # shuffle the data
        shuffled_ids = self.random_number_generator.permutation(self.size)

        for start in range(0, self.size, self.mini_batch_size):
            end = min(start + self.mini_batch_size, self.size)
            batch_ids = shuffled_ids[start:end]

            batch_data = {}
            for key in self.all_keys:
                batch_data[key] = self.data[key][batch_ids]

            yield batch_data

    def save_checkpoint(self, context: SaveContext):
        save_path = os.path.join(
            context.save_path, context.prefix + self.save_file_name
        )

        if not self.data:
            # Data is empty, save an file only containing an indicator
            to_save = self.FILE_EMPTY_SIGNAL
        else:
            to_save = self.data

        context.file_handler.save_to_file(to_save=to_save, save_path=save_path)

    def load_checkpoint(self, context: SaveContext):
        load_path = os.path.join(
            context.save_path, context.prefix + self.save_file_name
        )

        data = context.file_handler.load_from_file(load_path)
        if data == self.FILE_EMPTY_SIGNAL:
            # Data is empty, set an empty dictionary
            self.data = {}
            self.all_keys = []
            self.size = 0
        else:
            self.set_data(data)



from typing import Dict, Any, Tuple

from gymnasium.spaces import Space, Box, Discrete
import torch

import athlete
from athlete.update.update_rule import UpdateRule
from athlete.algorithms.dqn.update import DQNUpdate
from athlete.policy.policy_builder import PolicyBuilder
from athlete.algorithms.dqn.policy import DQNPolicyBuilder
from athlete.data_collection.provider import UpdateDataProvider
from athlete.module.torch.common import FCDiscreteQValueFunction
from athlete import constants
from athlete.data_collection.collector import DataCollector
from athlete.data_collection.transition import GymnasiumTransitionDataCollector

# Deep Q-Network

ARGUMENT_DISCOUNT = "discount"
ARGUMENT_VALUE_NETWORK_CLASS = "value_network_class"
ARGUMENT_VALUE_NETWORK_ARGUMENTS = "value_network_arguments"
ARGUMENT_OPTIMIZER_CLASS = "optimizer_class"
ARGUMENT_OPTIMIZER_ARGUMENTS = "optimizer_arguments"
ARGUMENT_REPLAY_BUFFER_CAPACITY = "replay_buffer_capacity"
ARGUMENT_REPLAY_BUFFER_MINI_BATCH_SIZE = "replay_buffer_mini_batch_size"
ARGUMENT_START_EPSILON = "start_epsilon"
ARGUMENT_END_EPSILON = "end_epsilon"
ARGUMENT_EPSILON_DECAY_STEPS = "epsilon_decay_steps"
ARGUMENT_VALUE_NET_UPDATE_FREQUENCY = "value_net_update_frequency"
ARGUMENT_VALUE_NET_NUMBER_OF_UPDATES = "value_net_number_of_updates"
ARGUMENT_MULTIPLY_NUMBER_OF_UPDATES_BY_ENVIRONMENT_STEPS = (
    "multiply_number_of_updates_by_environment_steps"
)
ARGUMENT_TARGET_NET_UPDATE_FREQUENCY = "target_net_update_frequency"
ARGUMENT_TARGET_NET_TAU = "target_net_tau"
ARGUMENT_ENABLE_DOUBLE_Q_LEARNING = "enable_double_q_learning"
ARGUMENT_CRITERIA = "criteria"
ARGUMENT_GRADIENT_MAX_NORM = "gradient_max_norm"
ARGUMENT_DEVICE = "device"
ARGUMENT_ADDITIONAL_REPLAY_BUFFER_ARGUMENTS = "additional_replay_buffer_arguments"
ARGUMENT_POST_REPLAY_BUFFER_DATA_PREPROCESSING = "post_replay_buffer_data_preprocessing"

DEFAULT_CONFIGURATION = {
    ARGUMENT_DISCOUNT: 0.99,
    ARGUMENT_VALUE_NETWORK_CLASS: FCDiscreteQValueFunction,
    ARGUMENT_VALUE_NETWORK_ARGUMENTS: {
        constants.GENERAL_ARGUMENT_OBSERVATION_SPACE: constants.VALUE_PLACEHOLDER,
        constants.GENERAL_ARGUMENT_ACTION_SPACE: constants.VALUE_PLACEHOLDER,
        "hidden_dims": [256, 256],
        "init_state_dict_path": None,
    },
    ARGUMENT_OPTIMIZER_CLASS: torch.optim.Adam,
    ARGUMENT_OPTIMIZER_ARGUMENTS: {"lr": 6.3e-4},
    ARGUMENT_REPLAY_BUFFER_CAPACITY: 100000,
    ARGUMENT_REPLAY_BUFFER_MINI_BATCH_SIZE: 128,
    ARGUMENT_START_EPSILON: 1.0,
    ARGUMENT_END_EPSILON: 0.1,
    constants.GENERAL_ARGUMENT_WARMUP_STEPS: 1000,
    ARGUMENT_EPSILON_DECAY_STEPS: 12000,
    ARGUMENT_VALUE_NET_UPDATE_FREQUENCY: 4,
    ARGUMENT_VALUE_NET_NUMBER_OF_UPDATES: 1,
    ARGUMENT_MULTIPLY_NUMBER_OF_UPDATES_BY_ENVIRONMENT_STEPS: True,
    ARGUMENT_TARGET_NET_UPDATE_FREQUENCY: 250,
    ARGUMENT_TARGET_NET_TAU: None,
    ARGUMENT_ENABLE_DOUBLE_Q_LEARNING: False,
    ARGUMENT_CRITERIA: torch.nn.MSELoss(),
    ARGUMENT_GRADIENT_MAX_NORM: None,
    ARGUMENT_DEVICE: (
        "cuda"
        if torch.cuda.is_available()
        else "mps" if torch.backends.mps.is_available() else "cpu"
    ),
    ARGUMENT_ADDITIONAL_REPLAY_BUFFER_ARGUMENTS: {},
    ARGUMENT_POST_REPLAY_BUFFER_DATA_PREPROCESSING: None,
}


def make_dqn_components(
    observation_space: Space, action_space: Space, configuration: Dict[str, Any]
) -> Tuple[DataCollector, UpdateRule, PolicyBuilder]:
    """Creates the components for a DQN agent.

    Args:
        observation_space: Environment observation space
        action_space: Environment action space
        configuration: Algorithm configuration dictionary

    Returns:
        Tuple containing data collector, update rule and policy builder

    Raises:
        ValueError: If observation_space is not Box or action_space is not Discrete
    """

    if not isinstance(observation_space, Box):
        raise ValueError(
            f"This DQN implementation only supports {Box.__name__} observation spaces, but got {type(observation_space)}"
        )
    if not isinstance(action_space, Discrete):
        raise ValueError(
            f"This DQN implementation only supports {Discrete.__name__} action spaces, but got {type(action_space)}"
        )

    configuration[ARGUMENT_VALUE_NETWORK_ARGUMENTS].update(
        {
            constants.GENERAL_ARGUMENT_OBSERVATION_SPACE: observation_space,
            constants.GENERAL_ARGUMENT_ACTION_SPACE: action_space,
        }
    )

    value_function = configuration[ARGUMENT_VALUE_NETWORK_CLASS](
        **configuration[ARGUMENT_VALUE_NETWORK_ARGUMENTS]
    )

    update_data_input = UpdateDataProvider()

    # DATA COLLECTOR

    data_collector = GymnasiumTransitionDataCollector(
        update_data_provider=update_data_input,
    )

    # UPDATE RULE
    update_rule = DQNUpdate(
        observation_space=observation_space,
        action_space=action_space,
        update_data_input=update_data_input,
        q_value_function=value_function,
        discount=configuration[ARGUMENT_DISCOUNT],
        optimizer_arguments=configuration[ARGUMENT_OPTIMIZER_ARGUMENTS],
        replay_buffer_capacity=configuration[ARGUMENT_REPLAY_BUFFER_CAPACITY],
        replay_buffer_mini_batch_size=configuration[
            ARGUMENT_REPLAY_BUFFER_MINI_BATCH_SIZE
        ],
        value_net_update_frequency=configuration[ARGUMENT_VALUE_NET_UPDATE_FREQUENCY],
        value_net_number_of_updates=configuration[ARGUMENT_VALUE_NET_NUMBER_OF_UPDATES],
        multiply_number_of_updates_by_environment_steps=configuration[
            ARGUMENT_MULTIPLY_NUMBER_OF_UPDATES_BY_ENVIRONMENT_STEPS
        ],
        target_net_update_frequency=configuration[ARGUMENT_TARGET_NET_UPDATE_FREQUENCY],
        target_net_tau=configuration[ARGUMENT_TARGET_NET_TAU],
        enable_double_q_learning=configuration[ARGUMENT_ENABLE_DOUBLE_Q_LEARNING],
        criteria=configuration[ARGUMENT_CRITERIA],
        optimizer_class=configuration[ARGUMENT_OPTIMIZER_CLASS],
        gradient_max_norm=configuration[ARGUMENT_GRADIENT_MAX_NORM],
        device=configuration[ARGUMENT_DEVICE],
        additional_replay_buffer_arguments=configuration[
            ARGUMENT_ADDITIONAL_REPLAY_BUFFER_ARGUMENTS
        ],
        post_replay_buffer_data_preprocessing=configuration[
            ARGUMENT_POST_REPLAY_BUFFER_DATA_PREPROCESSING
        ],
    )

    # POLICY BUILDER

    policy_builder = DQNPolicyBuilder(
        q_value_function=update_rule.q_value_function,
        action_space=action_space,
        start_epsilon=configuration[ARGUMENT_START_EPSILON],
        end_epsilon=configuration[ARGUMENT_END_EPSILON],
        epsilon_decay_steps=configuration[ARGUMENT_EPSILON_DECAY_STEPS],
        post_replay_buffer_preprocessing=configuration[
            ARGUMENT_POST_REPLAY_BUFFER_DATA_PREPROCESSING
        ],
    )

    return data_collector, update_rule, policy_builder


athlete.register(
    id="dqn",
    component_factory=make_dqn_components,
    default_configuration=DEFAULT_CONFIGURATION,
)



from athlete.agent import Agent
from athlete.algorithms.registry import AlgorithmRegistry

# This is done such that we can call these functions directly with athlete.<function>
make = Agent.make
from_checkpoint = Agent.from_checkpoint
get_default_configuration = AlgorithmRegistry.get_default_configuration
register = AlgorithmRegistry.register
list_algorithms = AlgorithmRegistry.list_algorithms

# Import all algorithms to register them, registration happens in according __init__.py files
import athlete.algorithms.q_learning
import athlete.algorithms.dqn
import athlete.algorithms.ddpg
import athlete.algorithms.td3
import athlete.algorithms.sac
import athlete.algorithms.ppo



from typing import Dict, Callable, Optional

import torch
from torch.optim import Optimizer

from athlete import constants
from athlete.update.common import TorchFrequentGradientUpdate


class TD3CriticUpdate(TorchFrequentGradientUpdate):
    """The updatable component for the TD3 critic update."""

    CRITIC_LOSS_LOG_TAG = "critic_loss"

    SAVE_HANDLING_STATS = "td3_value_update_handling_stats"

    def __init__(
        self,
        data_sampler: Callable[[None], Dict[str, torch.tensor]],
        critic_1: torch.nn.Module,
        critic_2: torch.nn.Module,
        target_critic_1: torch.nn.Module,
        target_critic_2: torch.nn.Module,
        critic_optimizer: Optimizer,
        target_actor: torch.nn.Module,
        discount: float = 0.99,
        target_noise_std: float = 0.2,
        target_noise_clip: float = 0.5,
        changes_policy: bool = False,
        update_frequency: int = 1,
        number_of_updates: int = 1,
        multiply_number_of_updates_by_environment_steps: bool = False,
        criteria: torch.nn.modules.loss._Loss = torch.nn.MSELoss(),
        log_tag: str = CRITIC_LOSS_LOG_TAG,
        save_file_name: str = SAVE_HANDLING_STATS,
        gradient_max_norm: Optional[float] = None,
    ) -> None:
        """Initializes the TD3 critic update.

        Args:
            data_sampler (Callable[[None], Dict[str, torch.tensor]]): A function that upon calling it returns a batch of data to use for updates in the form of a dictionary.
            critic_1 (torch.nn.Module): The first critic network to update.
            critic_2 (torch.nn.Module): The second critic network to update.
            target_critic_1 (torch.nn.Module): The target network for the first critic.
            target_critic_2 (torch.nn.Module): The target network for the second critic.
            critic_optimizer (Optimizer): The optimizer to use for both critic networks.
            target_actor (torch.nn.Module): The target actor network to use for the target value calculation.
            discount (float, optional): The discount factor for the value update. Defaults to 0.99.
            target_noise_std (float, optional): The standard deviation of the noise to add to the target actions. Defaults to 0.2.
            target_noise_clip (float, optional): The maximum absolute value of the noise to add to the target actions. Defaults to 0.5.
            changes_policy (bool, optional): Whether this update immediately changes the policy. For regular TD3 this is False. Defaults to False.
            update_frequency (int, optional): The update frequency according to the number of environment steps. If -1 an update is performed at the end of each episode. Defaults to 1.
            number_of_updates (int, optional): The number of updates to perform at each update step. Defaults to 1.
            multiply_number_of_updates_by_environment_steps (bool, optional): Whether to multiply the number of updates by the number of environment steps since the last update. Defaults to False.
            criteria (torch.nn.modules.loss._Loss, optional): The loss function used for calculating the loss between the critic predictions and the target values. Defaults to torch.nn.MSELoss().
            log_tag (str, optional): The log tag to use for logging the loss. Defaults to "critic_loss".
            save_file_name (str, optional): The file name to use for saving the handling stats. Defaults to "td3_value_update_handling_stats".
            gradient_max_norm (Optional[float], optional): The maximum norm for gradient clipping. If None, no clipping is performed. Defaults to None.
        """
        super().__init__(
            optimizer=critic_optimizer,
            changes_policy=changes_policy,
            log_tag=log_tag,
            save_file_name=save_file_name,
            update_frequency=update_frequency,
            number_of_updates=number_of_updates,
            multiply_number_of_updates_by_environment_steps=multiply_number_of_updates_by_environment_steps,
            gradient_max_norm=gradient_max_norm,
        )

        self.critic_1 = critic_1
        self.target_critic_1 = target_critic_1
        self.critic_2 = critic_2
        self.target_critic_2 = target_critic_2
        self.target_actor = target_actor
        self.target_noise_std = target_noise_std
        self.target_noise_clip = target_noise_clip
        self.discount = discount
        self.criteria = criteria
        self.data_sampler = data_sampler

    def calculate_loss(
        self,
    ) -> torch.Tensor:
        """Calculates the critics loss according to the TD3 algorithm.

        Returns:
            torch.Tensor: The loss of the critics.
        """

        batch_dictionary = self.data_sampler()

        observations = batch_dictionary[constants.DATA_OBSERVATIONS]
        actions = batch_dictionary[constants.DATA_ACTIONS]
        rewards = batch_dictionary[constants.DATA_REWARDS]
        next_observations = batch_dictionary[constants.DATA_NEXT_OBSERVATIONS]
        terminateds = batch_dictionary[constants.DATA_TERMINATEDS]

        # prediction
        prediction_1 = self.critic_1(observations, actions)
        prediction_2 = self.critic_2(observations, actions)

        # target
        with torch.no_grad():
            target = self._calculate_target(
                rewards=rewards,
                next_observations=next_observations,
                terminateds=terminateds,
            )

        loss_1 = self.criteria(prediction_1, target)
        loss_2 = self.criteria(prediction_2, target)

        return loss_1 + loss_2

    def _calculate_target(
        self,
        rewards: torch.tensor,
        next_observations: torch.tensor,
        terminateds: torch.tensor,
    ) -> torch.Tensor:
        """Calculates the target value for the critics according to the TD3 algorithm.

        Args:
            rewards (torch.tensor): Rewards batch of shape (batch_size, 1).
            next_observations (torch.tensor): Next observations batch of shape (batch_size, observation_size).
            terminateds (torch.tensor): Terminateds batch of shape (batch_size, 1).

        Returns:
            torch.Tensor: Target values of shape (batch_size, 1).
        """
        target = torch.clone(rewards)
        not_terminateds = ~terminateds.type(torch.bool).reshape(-1)
        next_actions = self.target_actor(next_observations[not_terminateds])
        # according to TD3 paper, we add clipped noise to the target actions
        clipped_noise = torch.clip(
            torch.randn_like(next_actions) * self.target_noise_std,
            min=-self.target_noise_clip,
            max=self.target_noise_clip,
        )
        # We assume unscaled actions here, scaling is done in policy and unscaled actions are used for update
        noisy_next_actions = torch.clip(next_actions + clipped_noise, min=-1.0, max=1.0)

        next_q_values_1 = self.target_critic_1(
            next_observations[not_terminateds],
            noisy_next_actions,
        )
        next_q_values_2 = self.target_critic_2(
            next_observations[not_terminateds],
            noisy_next_actions,
        )
        min_next_q_values = torch.min(next_q_values_1, next_q_values_2)
        target[not_terminateds] += self.discount * min_next_q_values
        return target



from typing import Protocol, Dict, Any, runtime_checkable, Type, Callable
import os
from dataclasses import dataclass

import torch
import numpy as np

from athlete.saving.file_handler import FileHandler


@dataclass
class SaveContext:
    """Dataclass that provides context information for saving and loading operations.

    This class encapsulates all necessary information required by components during
    checkpoint operations, including file handling, paths, naming prefixes, and metadata.
    Prefixes can be used if the same object type is used in different places to avoid name clashes.
    """

    file_handler: FileHandler
    save_path: str
    prefix: str
    metadata: Dict[str, Any] = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}

    def with_prefix(self, additional_prefix: str) -> "SaveContext":
        """Create a new context with an additional prefix.

        Args:
            additional_prefix (str): Additional prefix to append to the existing prefix.

        Returns:
            SaveContext: New SaveContext with the updated prefix.
        """
        return SaveContext(
            file_handler=self.file_handler,
            save_path=self.save_path,
            prefix=f"{self.prefix}{additional_prefix}",
            metadata=self.metadata.copy(),
        )

    def replace_metadata(self, **kwargs) -> "SaveContext":
        """Create a new context with replaced metadata.

        Returns:
            SaveContext: New SaveContext with the replaced metadata.
        """
        return SaveContext(
            file_handler=self.file_handler,
            save_path=self.save_path,
            prefix=self.prefix,
            metadata=kwargs,
        )


@runtime_checkable
class SaveableComponent(Protocol):
    """Protocol for saveable components. Any object that implements this protocol can be saved and loaded."""

    def save_checkpoint(self, context: SaveContext) -> None:
        """Save this object's state using the provided context.

        Args:
            context (SaveContext): The context to use for saving the object's state.
        """
        pass

    def load_checkpoint(self, context: SaveContext) -> None:
        """Load this object's state using the provided context.

        Args:
            context (SaveContext): The context to use for loading the object's state.
        """
        pass


class CompositeSaveableComponent:
    """An object that contains multiple saveable components. It provides functionality to register
    saveable components such that they will be saved and loaded upon calling the respective methods.
    The components can be either an object following the SaveableComponent Protocol or a saver function must be registered
    in the SaverRegistry for this type.
    """

    def __init__(self) -> None:
        """Initialize the CompositeSaveableComponent.
        This initializes two dictionaries to keep track of the components:
        - _saveable_components: Components that implement the SaveableComponent Protocol
        - _external_components: Components that can not save themselves but need a saver function registered in the SaverRegistry
        """
        self._saveable_components = {}  # Components that implement Saveable
        self._external_components = {}  # Components that need SaverRegistry

    def register_saveable_component(
        self,
        name: str,
        component: Any,
    ) -> None:
        """Register a component under the given name.

        The component is automatically categorized as either saveable or external.
        For externals, this checks if a matching saver exists in the registry.

        Args:
            name: Name to identify this component, also used as the filename for external components
            component: The component to save

        Raises:
            TypeError: If component is not saveable and no saver is registered for its type.
        """

        # Check if component implements the Saveable protocol by looking for required methods
        is_saveable = isinstance(component, SaveableComponent)

        if is_saveable:
            self._saveable_components[name] = component
        else:
            # Check if there's a registered saver for this component type
            has_saver = False
            for data_type in SaverRegistry._savers.keys():
                # This also considers subclassing
                if isinstance(component, data_type):
                    has_saver = True
                    break

            if not has_saver:
                raise TypeError(
                    f"Cannot register component '{name}' of type {type(component).__name__}: "
                    f"it does not implement the saveable component protocol and no saver is registered for this type."
                )

            self._external_components[name] = component

    def save_checkpoint(self, context: SaveContext) -> None:
        """Save all registered components.
        And save any local state according to the _save_local_state function.

        Args:
            context (SaveContext): The context to use for saving the object's state.
        """
        # First save any local state
        self._save_local_state(context)

        # Save all Saveable components
        for name, component in self._saveable_components.items():
            component.save_checkpoint(context=context)

        # Save all external components using SaverRegistry
        for name, component in self._external_components.items():
            SaverRegistry.save(obj=component, context=context, filename=name)

    def load_checkpoint(self, context: SaveContext) -> None:
        """Load all registered components.
        And load any local state according to the _load_local_state function.

        Args:
            context (SaveContext): The context to use for loading the object's state.

        Raises:
            RuntimeError: If loading a external component fails.
        """
        # First load any local state
        self._load_local_state(context)

        # Load all Saveable components
        for name, component in self._saveable_components.items():
            component.load_checkpoint(context)

        # Load all external components using SaverRegistry
        for name, component in self._external_components.items():
            try:
                SaverRegistry.load(component, context, name)
            except Exception as e:
                raise RuntimeError(
                    f"Failed to load external component '{name}': {e}"
                ) from e

    def _save_local_state(self, context: SaveContext) -> None:
        """Override to save component-specific state.

        Args:
            context (SaveContext): The context to use for saving the object's state.
        """
        pass

    def _load_local_state(self, context: SaveContext) -> None:
        """Override to load component-specific state.

        Args:
            context (SaveContext): The context to use for loading the object's state.
        """
        pass


class SaverRegistry:
    """A registry of specialized saving and loading function for different data types.
    This is useful of the object you want to save can not implement the SaveableComponent protocol.
    """

    _savers = {}
    _loaders = {}

    @classmethod
    def register(
        cls,
        data_type: Type,
        saver_fn: Callable[[Any, SaveContext, str], None],
        loader_fn: Callable[[Any, SaveContext, str], None],
    ):
        """Register a saver and loader function for a specific data type.

        Args:
            data_type (Type): Data type to register the saver and loader for
            saver_fn (Callable[[Any, SaveContext, str], None]): Saver function to register
            loader_fn (Callable[[Any, SaveContext, str], None]): Loader function to register,
            note that this returns none, it has to overwrite the object in place.
        """
        cls._savers[data_type] = saver_fn
        cls._loaders[data_type] = loader_fn

    @classmethod
    def save(cls, obj: Any, context: SaveContext, filename: str) -> None:
        """Save an object using the appropriate registered saver.

        Args:
            obj (_type_): Object to save
            context (SaveContext): The save context
            filename (str): Name of the file to save to

        Raises:
            TypeError: If no registered saver is found for the object's type
        """
        for data_type, saver_fn in cls._savers.items():
            if isinstance(obj, data_type):
                saver_fn(obj, context, filename)
                return
        raise TypeError(f"No registered saver for object of type: {type(obj)}")

    @classmethod
    def load(cls, obj: Any, context: SaveContext, filename: str) -> None:
        """Load an object using the appropriate registered loader.
        This function modifies the object in place and returns None.

        Args:
            obj (Any): Object to load
            context (SaveContext): The save context
            filename (str): Name of the file to load from

        Raises:
            TypeError: If no registered loader is found for the object's type
        """
        for data_type, loader_fn in cls._loaders.items():
            if isinstance(obj, data_type):
                loader_fn(obj, context, filename)
                return

        raise TypeError(f"No registered loader for type: {obj}")


# Register specialized savers, note that the load functions need to modify the to_load object in place


def save_torch_module(
    to_save: torch.nn.Module, context: SaveContext, filename: str
) -> None:
    """Save a PyTorch module."""
    save_path = os.path.join(context.save_path, context.prefix + filename)
    context.file_handler.save_to_file(to_save.state_dict(), save_path)


def load_torch_module(
    to_load: torch.nn.Module, context: SaveContext, filename: str
) -> None:
    """Load a PyTorch module."""
    load_path = os.path.join(context.save_path, context.prefix + filename)
    state_dict = context.file_handler.load_from_file(load_path)
    to_load.load_state_dict(state_dict)
    return to_load


def save_torch_optimizer(
    to_save: torch.optim.Optimizer, context: SaveContext, filename: str
) -> None:
    """Save a PyTorch optimizer."""
    save_path = os.path.join(context.save_path, context.prefix + filename)
    context.file_handler.save_to_file(to_save.state_dict(), save_path)


def load_torch_optimizer(
    to_load: torch.optim.Optimizer, context: SaveContext, filename: str
) -> None:
    """Load a PyTorch optimizer."""
    load_path = os.path.join(context.save_path, context.prefix + filename)
    state_dict = context.file_handler.load_from_file(load_path)
    to_load.load_state_dict(state_dict)


def save_default(to_save: Any, context: SaveContext, filename: str) -> None:
    """Saves anything that doesn't require special handling."""
    save_path = os.path.join(context.save_path, context.prefix + filename)
    context.file_handler.save_to_file(to_save, save_path)


def load_numpy_array(to_load: np.ndarray, context: SaveContext, filename: str) -> None:
    """Load a NumPy array."""
    load_path = os.path.join(context.save_path, context.prefix + filename)
    loaded = context.file_handler.load_from_file(load_path)
    np.copyto(to_load, loaded)


def save_torch_parameter(
    to_save: torch.nn.Parameter, context: SaveContext, filename: str
) -> None:
    """Save a PyTorch parameter."""
    save_path = os.path.join(context.save_path, context.prefix + filename)
    context.file_handler.save_to_file(to_save.data, save_path)


def load_torch_parameter(
    to_load: torch.nn.Parameter, context: SaveContext, filename: str
) -> None:
    """Load a PyTorch parameter."""
    load_path = os.path.join(context.save_path, context.prefix + filename)
    loaded = context.file_handler.load_from_file(load_path)
    to_load.data.copy_(loaded)


# Register the savers
SaverRegistry.register(torch.nn.Module, save_torch_module, load_torch_module)
SaverRegistry.register(
    torch.optim.Optimizer, save_torch_optimizer, load_torch_optimizer
)
SaverRegistry.register(np.ndarray, save_default, load_numpy_array)
SaverRegistry.register(torch.nn.Parameter, save_torch_parameter, load_torch_parameter)



import torch
from typing import Dict, Any, Callable

from athlete import constants
from athlete.update.common import TorchFrequentGradientUpdate
from athlete.algorithms.sac.module import SACActor


class SACCriticUpdate(TorchFrequentGradientUpdate):
    """The updatable component for the critic of the SAC algorithm. It updates the two critics"""

    CRITIC_LOSS_LOG_TAG = "critic_loss"

    SAVE_HANDLING_STATS = "sac_value_update_handling_stats"

    def __init__(
        self,
        temperature: torch.nn.Parameter,
        actor: SACActor,
        critic_1: torch.nn.Module,
        critic_2: torch.nn.Module,
        critic_optimizer: torch.optim.Optimizer,
        target_critic_1: torch.nn.Module,
        target_critic_2: torch.nn.Module,
        data_sampler: Callable[[None], Dict[str, torch.Tensor]],
        discount: float = 0.99,
        criteria: torch.nn.modules.loss._Loss = torch.nn.MSELoss(),
        update_frequency: int = 1,
        number_of_updates: int = 1,
        multiply_number_of_updates_by_environment_steps: bool = False,
        changes_policy: bool = False,
        gradient_max_norm: float = None,
        log_tag: str = CRITIC_LOSS_LOG_TAG,
        save_file_name: str = SAVE_HANDLING_STATS,
    ) -> None:
        """Initializes the SACCriticUpdate class.

        Args:
            temperature (torch.nn.Parameter): The temperature parameter for the SAC algorithm.
            actor (SACActor): The actor of the SAC algorithm.
            critic_1 (torch.nn.Module): The first critic to update.
            critic_2 (torch.nn.Module): The second critic to update.
            critic_optimizer (torch.optim.Optimizer): The optimizer for both critics.
            target_critic_1 (torch.nn.Module): Target network for the first critic.
            target_critic_2 (torch.nn.Module): Target network for the second critic.
            data_sampler (Callable[[None], Dict[str, torch.Tensor]]): A function that returns a batch of data to update on upon call.
            discount (float, optional): The discount factor for the SAC algorithm. Defaults to 0.99.
            criteria (torch.nn.modules.loss._Loss, optional): Function used to calculate the loss between the target and the predicted Q-values. Defaults to torch.nn.MSELoss().
            update_frequency (int, optional): The frequency in which this updated is applied according to the environment steps. If -1, it is applied at the end of each episode. Defaults to 1.
            number_of_updates (int, optional): The number of updates to perform per call. Defaults to 1.
            multiply_number_of_updates_by_environment_steps (bool, optional): Whether to multiply the number of updates by the environment steps since the last update. Defaults to False.
            changes_policy (bool, optional): Whether applying this update immediately changes the policy, for regular SAC this is False. Defaults to False.
            gradient_max_norm (float, optional): The maximum norm for the gradients. If None, no gradient clipping is applied. Defaults to None.
            log_tag (str, optional): Log tag for the loss. Defaults to "critic_loss".
            save_file_name (str, optional): File name for saving handling stats. Defaults to "sac_value_update_handling_stats".
        """
        super().__init__(
            optimizer=critic_optimizer,
            changes_policy=changes_policy,
            log_tag=log_tag,
            save_file_name=save_file_name,
            update_frequency=update_frequency,
            number_of_updates=number_of_updates,
            multiply_number_of_updates_by_environment_steps=multiply_number_of_updates_by_environment_steps,
            gradient_max_norm=gradient_max_norm,
        )

        self.temperature = temperature
        self.actor = actor
        self.critic_1 = critic_1
        self.critic_2 = critic_2
        self.target_critic_1 = target_critic_1
        self.target_critic_2 = target_critic_2
        self.data_sampler = data_sampler
        self.discount = discount
        self.criteria = criteria

    def calculate_loss(
        self,
    ) -> torch.Tensor:
        """Calculates the critic loss according to the SAC algorithm.

        Returns:
            torch.Tensor: The loss of the critics.
        """

        batch_dictionary = self.data_sampler()

        observations = batch_dictionary[constants.DATA_OBSERVATIONS]
        actions = batch_dictionary[constants.DATA_ACTIONS]
        rewards = batch_dictionary[constants.DATA_REWARDS]
        next_observations = batch_dictionary[constants.DATA_NEXT_OBSERVATIONS]
        terminateds = batch_dictionary[constants.DATA_TERMINATEDS]

        not_terminateds = ~terminateds.type(torch.bool).reshape(-1)

        with torch.no_grad():
            next_actions, next_log_probabilities = self.actor.get_action_and_log_prob(
                next_observations[not_terminateds]
            )
            next_q_values_1 = self.target_critic_1(
                next_observations[not_terminateds], next_actions
            )
            next_q_values_2 = self.target_critic_2(
                next_observations[not_terminateds], next_actions
            )
            min_next_q_values = torch.min(next_q_values_1, next_q_values_2)
            min_next_q_values_with_entropy = (
                min_next_q_values - self.temperature * next_log_probabilities
            )
            target_q_values = torch.clone(rewards)
            target_q_values[not_terminateds] += (
                self.discount * min_next_q_values_with_entropy
            )

        q_value_predictions_1 = self.critic_1(observations, actions)
        q_value_predictions_2 = self.critic_2(observations, actions)
        loss_1 = self.criteria(q_value_predictions_1, target_q_values)
        loss_2 = self.criteria(q_value_predictions_2, target_q_values)
        return loss_1 + loss_2


class SACActorUpdate(TorchFrequentGradientUpdate):
    """The updatable component for the actor of the SAC algorithm. It updates the actor network."""

    ACTOR_LOSS_LOG_TAG = "actor_loss"

    SAVE_HANDLING_STATS = "sac_actor_update_handling_stats"

    def __init__(
        self,
        actor: SACActor,
        actor_optimizer: torch.optim.Optimizer,
        critic_1: torch.nn.Module,
        critic_2: torch.nn.Module,
        temperature: torch.nn.Parameter,
        data_sampler: Callable[[None], Dict[str, torch.Tensor]],
        update_frequency: int = 1,
        number_of_updates: int = 1,
        multiply_number_of_updates_by_environment_steps: bool = False,
        gradient_max_norm: float = None,
        changes_policy: bool = True,
        log_tag: str = ACTOR_LOSS_LOG_TAG,
        save_file_name: str = SAVE_HANDLING_STATS,
    ) -> None:
        """Initializes the SAC actor update class.

        Args:
            actor (SACActor): The actor to update.
            actor_optimizer (torch.optim.Optimizer): The optimizer of the actor.
            critic_1 (torch.nn.Module): The first critic to use for the update.
            critic_2 (torch.nn.Module): The second critic to use for the update.
            temperature (torch.nn.Parameter): The temperature parameter for the SAC algorithm.
            data_sampler (Callable[[None], Dict[str, torch.Tensor]]): A function that returns a batch of data to update on upon call.
                Here, the batch only needs to contain the observations.
            update_frequency (int, optional): The frequency in which this updated is applied according to the environment steps. If -1, it is applied at the end of each episode. Defaults to 1.
            number_of_updates (int, optional): The number of updates to perform per call. Defaults to 1.
            multiply_number_of_updates_by_environment_steps (bool, optional): Whether to multiply the number of updates by the environment steps since the last update. Defaults to False.
            gradient_max_norm (float, optional): The maximum norm for the gradients. If None, no gradient clipping is applied. Defaults to None.
            changes_policy (bool, optional): Whether applying this update immediately changes the policy, for regular SAC this is True. Defaults to True.
            log_tag (str, optional): Log tag for the loss. Defaults to "actor_loss".
            save_file_name (str, optional): File name for saving handling stats. Defaults to "sac_actor_update_handling_stats".
        """
        super().__init__(
            optimizer=actor_optimizer,
            changes_policy=changes_policy,
            log_tag=log_tag,
            save_file_name=save_file_name,
            update_frequency=update_frequency,
            number_of_updates=number_of_updates,
            multiply_number_of_updates_by_environment_steps=multiply_number_of_updates_by_environment_steps,
            gradient_max_norm=gradient_max_norm,
        )

        self.actor = actor
        self.critic_1 = critic_1
        self.critic_2 = critic_2
        self.temperature = temperature
        self.data_sampler = data_sampler

    def calculate_loss(
        self,
    ) -> torch.Tensor:
        """Calculates the actor loss according to the SAC algorithm.

        Returns:
            torch.Tensor: The loss of the actor.
        """
        batch_dictionary = self.data_sampler()

        observations = batch_dictionary[constants.DATA_OBSERVATIONS]

        actions, log_probabilities = self.actor.get_action_and_log_prob(observations)
        q_values_1 = self.critic_1(observations, actions)
        q_values_2 = self.critic_2(observations, actions)
        min_q_values = torch.min(q_values_1, q_values_2)
        loss = torch.mean((self.temperature * log_probabilities) - min_q_values)
        return loss


class SACTemperatureUpdate(TorchFrequentGradientUpdate):
    """The updatable component for the temperature of the SAC algorithm. It updates the temperature parameter
    and can be used if the temperature should be adjusted automatically.
    """

    TEMPERATURE_LOSS_LOG_TAG = "temperature_loss"
    TEMPERATURE_LOG_TAG = "temperature"

    SAVE_HANDLING_STATS = "sac_temperature_update_handling_stats"

    def __init__(
        self,
        target_entropy: float,
        log_temperature: torch.nn.Parameter,
        temperature: torch.nn.Parameter,
        temperature_optimizer: torch.optim.Optimizer,
        actor: SACActor,
        data_sampler: Callable[[None], Dict[str, torch.Tensor]],
        update_frequency: int = 1,
        number_of_updates: int = 1,
        multiply_number_of_updates_by_environment_steps: bool = False,
        changes_policy: bool = False,
        gradient_max_norm: float = None,
        temperature_log_tag: str = TEMPERATURE_LOG_TAG,
        loss_log_tag: str = TEMPERATURE_LOSS_LOG_TAG,
        save_file_name: str = SAVE_HANDLING_STATS,
    ) -> None:
        """Initializes the SAC temperature update class.

        Args:
            target_entropy (float): The target entropy for the actor.
            log_temperature (torch.nn.Parameter): The log temperature parameter, this is updated.
            temperature (torch.nn.Parameter): The actual temperature parameter, this is updated by taking the exponential of the log temperature.
            temperature_optimizer (torch.optim.Optimizer): The optimizer for the temperature parameter.
            actor (SACActor): The stochastic actor of the SAC algorithm.
            data_sampler (Callable[[None], Dict[str, torch.Tensor]]): A function that returns a batch of data to update on upon call, here it only needs to return the observations.
            update_frequency (int, optional): The frequency in which this updated is applied according to the environment steps. If -1, it is applied at the end of each episode. Defaults to 1.
            number_of_updates (int, optional): The number of updates to perform per call. Defaults to 1.
            multiply_number_of_updates_by_environment_steps (bool, optional): Whether to multiply the number of updates by the environment steps since the last update. Defaults to False.
            changes_policy (bool, optional): Whether applying this update immediately changes the policy, for regular SAC this is False. Defaults to False.
            gradient_max_norm (float, optional): The maximum norm for the gradients. If None, no gradient clipping is applied. Defaults to None.
            temperature_log_tag (str, optional): Log tag used for the log temperature loss. Defaults to "temperature_loss".
            loss_log_tag (str, optional): Log tag used to log the current temperature. Defaults to "temperature".
            save_file_name (str, optional): File name for saving handling stats. Defaults to "sac_temperature_update_handling_stats".
        """
        super().__init__(
            optimizer=temperature_optimizer,
            changes_policy=changes_policy,
            log_tag=loss_log_tag,
            save_file_name=save_file_name,
            update_frequency=update_frequency,
            number_of_updates=number_of_updates,
            multiply_number_of_updates_by_environment_steps=multiply_number_of_updates_by_environment_steps,
            gradient_max_norm=gradient_max_norm,
        )

        self.target_entropy = target_entropy
        self.log_temperature = log_temperature
        self.temperature = temperature
        self.actor = actor
        self.data_sampler = data_sampler

        self.number_of_updates = number_of_updates
        self.update_per_environment_interaction = (
            multiply_number_of_updates_by_environment_steps
        )

        self.temperature_log_tag = temperature_log_tag
        self.loss_log_tag = loss_log_tag

    def calculate_loss(
        self,
    ) -> torch.Tensor:
        """Calculates the temperature loss according to the target entropy.

        Returns:
            torch.Tensor: The loss of the temperature.
        """
        batch_dictionary = self.data_sampler()
        observations = batch_dictionary[constants.DATA_OBSERVATIONS]

        with torch.no_grad():
            _, log_probabilities = self.actor.get_action_and_log_prob(observations)
        loss = torch.mean(
            -torch.exp(self.log_temperature) * (log_probabilities + self.target_entropy)
        )
        return loss

    def post_update_routine(self) -> Dict[str, Any]:
        """Updates the actual temperature as the exponential of the log temperature
        and add the current temperature to the logging dictionary.

        Returns:
            Dict[str, Any]: Dictionary containing the current temperature.
        """
        # Update actual temperature value used by other components
        self.temperature.data = torch.exp(self.log_temperature)
        return {self.temperature_log_tag: self.temperature.item()}



import os
from typing import Optional

import numpy as np
import torch

from athlete.saving.saveable_component import SaveContext


class StepTracker:
    """The StepTracker holds the number of interactions, episodes and datapoints added to the replay buffer.
    This information might be useful for updating conditions. This class can be used as a singleton.
    """

    FILE_SAVE_NAME = "step_tracker"

    _instance = None

    @classmethod
    def get_instance(cls) -> "StepTracker":
        """Returns the global instance of the StepTracker.

        Raises:
            Exception: If the instance has not been initialized.

        Returns:
            StepTracker: The global instance of the StepTracker.
        """
        if cls._instance is None:
            raise Exception("TrainingTracker has not been initialized.")
        return cls._instance

    @classmethod
    def set_global_instance(cls, instance: "StepTracker") -> None:
        """Sets the global instance of the StepTracker.

        Args:
            instance (StepTracker): The instance to set as the global instance.
        """
        cls._instance = instance

    def __init__(self, warmup_steps: int) -> None:
        """Initializes a StepTracker instance. Can be different from the global instance.

        Args:
            warmup_steps (int): Number of warmup steps that an algorithm might perform, this affects the interactions_after_warmup property.
        """
        self.total_interactions = 0
        self.total_number_of_datapoints_added = 0
        self.total_number_of_episodes = 0
        self.warmup_steps = warmup_steps

    def increment_environment_interaction(self, num_interactions: int = 1) -> None:
        """Increments the number of interactions with the environment.

        Args:
            num_interactions (int, optional): Number of interactions to add. Defaults to 1.
        """
        self.total_interactions += num_interactions

    def increment_datapoint(self, num_datapoints: int = 1) -> None:
        """Increments the number of collected datapoints, what exactly this means depends on the algorithm.

        Args:
            num_datapoints (int, optional): Number of datapoints to add. Defaults to 1.
        """
        self.total_number_of_datapoints_added += num_datapoints

    def increment_episode(self, num_episodes: int = 1) -> None:
        """Increments the number of episodes.

        Args:
            num_episodes (int, optional): Number of episodes to add. Defaults to 1.
        """
        self.total_number_of_episodes += num_episodes

    @property
    def interactions_after_warmup(self) -> int:
        """Returns the number of interactions after the warmup period.

        Returns:
            int: Number of interactions after the warmup period.
        """
        return max(0, self.total_interactions - self.warmup_steps)

    @property
    def warmup_is_done(self) -> bool:
        """Checks if the warmup period is done.

        Returns:
            bool: True if the warmup period is done, False otherwise.
        """
        return self.total_interactions >= self.warmup_steps

    def save_checkpoint(self, context: SaveContext) -> None:

        to_save = (
            self.total_interactions,
            self.total_number_of_episodes,
            self.total_number_of_datapoints_added,
        )
        save_path = os.path.join(
            context.save_path, context.prefix + self.FILE_SAVE_NAME
        )

        context.file_handler.save_to_file(to_save=to_save, save_path=save_path)

    def load_checkpoint(self, context: SaveContext) -> None:

        load_path = os.path.join(
            context.save_path, context.prefix + self.FILE_SAVE_NAME
        )

        (
            self.total_interactions,
            self.total_number_of_episodes,
            self.total_number_of_datapoints_added,
        ) = context.file_handler.load_from_file(load_path=load_path)


class RNGHandler:
    """This class handles random number generation to ensure reproducibility across training runs.
    It should be used as a singleton. It sets the global seed for numpy and torch as well as provides 
    a consistent random number generator instance from numpy that can be accessed throughout the codebase.
    """

    FILE_SAVE_NAME = "rng_handler"

    _instance = None

    @classmethod
    def get_random_number_generator(cls) -> np.random.Generator:
        """Returns the global random number generator instance.

        Raises:
            Exception: If the instance has not been initialized.

        Returns:
            np.random.Generator: The global random number generator instance.
        """
        if cls._instance is None:
            raise Exception("RNGHandler has not been initialized.")
        return cls._instance._random_number_generator

    @classmethod
    def get_seed(cls) -> int:
        """Returns the global seed.

        Raises:
            Exception: If the instance has not been initialized.

        Returns:
            int: The global seed.
        """
        if cls._instance is None:
            raise Exception("RNGHandler has not been initialized.")
        return cls._instance.seed

    @classmethod
    def get_instance(cls) -> "RNGHandler":
        """Returns the global instance of the RNGHandler.

        Raises:
            Exception: If the instance has not been initialized.

        Returns:
            RNGHandler: The global instance of the RNGHandler.
        """
        if cls._instance is None:
            raise Exception("RNGHandler has not been initialized.")
        return cls._instance

    @classmethod
    def set_global_instance(cls, rng_handler: "RNGHandler") -> None:
        """Sets the global instance of the RNGHandler.

        Args:
            rng_handler (RNGHandler): The instance to set as the global instance.
        """
        cls._instance = rng_handler

    def __init__(self, seed: Optional[int] = None) -> None:
        """Initializes a RNGHandler instance. This can be different from the global instance.

        Args:
            seed (Optional[int], optional): Seed for the random number generator. If None, a random seed will be generated. Defaults to None.
        """

        if not seed:
            seed = int(np.random.randint(low=0, high=np.iinfo(np.uint32).max))
        self.seed = seed

        # For convenience this ensures reproducibility for functions that do not use the global rng but numpy or torch
        # as long as they are used in the same order
        # Set seed for all numpy functions
        np.random.seed(seed)
        # Set seed for all torch functions
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

        # Create a random number generator instance
        # Best practice always use this one if possible
        self._random_number_generator = np.random.default_rng(seed)

    def save_checkpoint(self, context: SaveContext) -> None:

        # Global numpy random state
        np_state = np.random.get_state()

        # Global torch random state (CPU)
        torch_state = torch.get_rng_state()

        # Save CUDA states for all devices
        cuda_states = {}
        if torch.cuda.is_available():
            for i in range(torch.cuda.device_count()):
                cuda_states[i] = torch.cuda.get_rng_state(i)

        # Random number generator state
        rng_state = self._random_number_generator.bit_generator.state

        to_save = (self.seed, np_state, torch_state, cuda_states, rng_state)

        context.file_handler.save_to_file(
            to_save=to_save,
            save_path=os.path.join(
                context.save_path, context.prefix + self.FILE_SAVE_NAME
            ),
        )

    def load_checkpoint(self, context: SaveContext) -> None:

        loaded = context.file_handler.load_from_file(
            load_path=os.path.join(
                context.save_path, context.prefix + self.FILE_SAVE_NAME
            )
        )

        self.seed, np_state, torch_state, cuda_states, rng_state = loaded
        # Set global numpy random state
        np.random.set_state(np_state)

        # Set global torch random state (CPU)
        torch.set_rng_state(torch_state)

        # Set CUDA states for all devices
        if cuda_states and torch.cuda.is_available():
            for device_id, state in cuda_states.items():
                if device_id < torch.cuda.device_count():
                    torch.cuda.set_rng_state(state, device_id)

        # Set random number generator state
        self._random_number_generator.bit_generator.state = rng_state



from typing import Any, Dict, Optional, Tuple


class UpdateDataProvider:
    """A communication channel between data collectors and update rules.

    This container holds data which can be updated by data collectors and retrieved by update rules.
    It serves as a decoupling mechanism for transferring experience data between the collection
    phase and the learning phase without requiring direct dependencies between these components.
    """

    def __init__(
        self,
        initial_data: Optional[Any] = None,
        initial_metadata: Dict[str, Any] = {},
    ) -> None:
        """Initialize the UpdateDataProvider.

        Args:
            initial_data (Optional[Any], optional): Initial data held in the provider.
                Defaults to None.
            initial_metadata (Dict[str, Any], optional): _Initial metadata held in the provider.
                Defaults to {}.
        """
        self.update_data = initial_data
        self.metadata = initial_metadata

    def set_data(self, update_data: Any, metadata: Dict[str, Any] = {}) -> None:
        """Set new data in the provider.

        Args:
            update_data (Any): New data to be set in the provider.
            metadata (Dict[str, Any], optional): New metadata to be set in the provider.
        """
        self.update_data = update_data
        self.metadata = metadata

    def get_data(self) -> Tuple[Any, Dict[str, Any]]:
        """Get the current data and metadata from the provider.

        Returns:
            Tuple[Any, Dict[str, Any]]: The current data and metadata held in the provider.
        """
        return self.update_data, self.metadata



from typing import Dict, List, Any, Optional
import numpy as np
from abc import ABC, abstractmethod
import os
import sys
from collections import deque
from copy import deepcopy
import math
import gc

import cpprb

from athlete import constants
from athlete.saving.saveable_component import SaveContext
from athlete.global_objects import RNGHandler


class Buffer(ABC):
    """Base class for replay buffers commonly used in off-policy reinforcement learning algorithms.
    
    This abstract class defines the interface that all buffer implementations should follow,
    providing standard methods for adding transitions and sampling batches of experiences.
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def add(
        self,
        data_dictionary: Dict[str, np.ndarray],
        metadata: Optional[Dict[str, any]] = None,
    ) -> None:
        """Add data to the replay buffer.

        Args:
            data_dictionary (Dict[str, np.ndarray]): Data to be added to the replay buffer.
            metadata (Optional[Dict[str, any]], optional): Metadata that somehow affect the buffer,
            e.g. end episode signal used for memory compression for observation and next observation. Defaults to None.
        """
        ...

    @abstractmethod
    def sample(self, batch_size: int) -> Dict[str, np.ndarray]:
        """Sample data from the replay buffer.

        Args:
            batch_size (int): Number of entries to return.
        """
        ...

    @abstractmethod
    def encode_sample(self, sample_ids: List[int]) -> Dict[str, np.ndarray]:
        """Get specific data from the replay buffer by id.

        Args:
            sample_ids (List[int]): List of indices to get from the replay buffer.

        Returns:
            Dict[str, np.ndarray]: Retrieved data.
        """
        ...

    @property
    def size(self) -> int:
        """Current size of the replay buffer."""
        ...

    @property
    def max_size(self) -> int:
        """Maximum capacity of the replay buffer."""
        ...

    @property
    def data_info(self) -> Dict[str, Dict[str, Any]]:
        """Information describing the content of the replay buffer, (field names, shapes, dtypes)."""
        ...

    @property
    def pointer_position(self) -> int:
        """Current position of the pointer in the replay buffer (position where the next data will be added)."""
        ...

    def save_checkpoint(self, context: SaveContext):
        pass

    def load_checkpoint(self, context: SaveContext):
        pass


class EpisodicCPPReplayBuffer(Buffer):
    """Replay buffer using the cpprb library. This replay buffer can handle large amounts of data and use various data compression methods.
    It provides saving functionalities that allow for saving the replay buffer in chunks if it is too large to be saved in one go.
    """

    # based on https://ymd_h.gitlab.io/cpprb/
    # 4 GB in bytes
    REPLAY_SPLIT_SIZE = 4 * 1024 * 1024 * 1024

    FILE_REPLAY_BUFFER_HANDLING = "replay_buffer_handling_stats"
    FILE_REPLAY_BUFFER = "cpp_replay_buffer"
    FILE_EMPTY_SIGNAL = "empty"

    # Almost all of the complexity in this class stems from handling saves such that:
    # 1. The insertion point ends up in the correct place
    # 2. The episode borders are correctly saved and loaded which is relevant for compressing data in memory
    # 3. The replay buffer is saved in chunks if it is to large to be saved in one go

    def __init__(
        self,
        capacity: int,
        replay_buffer_info: Dict[str, Dict[str, Any]],
        additional_arguments: Dict[str, Any] = {},
    ) -> None:
        """Initialize the replay buffer.

        Args:
            capacity (int): Maximum capacity of the replay buffer.
            replay_buffer_info (Dict): Information describing the content of the replay buffer, (field names, shapes, dtypes). See cpprb documentation for more details. https://ymd_h.gitlab.io/cpprb/
            additional_arguments (Dict[str, Any], optional): Additional arguments passed to the cpprb initialization. Can be sued for specific memory compressions. Defaults to {}.
        """
        super().__init__()
        initialization_replay_buffer_info = (
            self._create_initialization_replay_buffer_info(
                replay_buffer_info=replay_buffer_info,
                additional_arguments=additional_arguments,
            )
        )
        self.replay_buffer = cpprb.ReplayBuffer(
            capacity, initialization_replay_buffer_info, **additional_arguments
        )
        self._max_size = capacity
        self.last_added_end_position = 0
        self.replay_buffer_info = replay_buffer_info
        # Useful to determine batch_size
        self.some_data_key = list(self.replay_buffer_info.keys())[0]

        self.buffer_episode_ended = np.full(
            self._max_size, dtype=np.bool_, fill_value=False
        )

        self.random_numbers_generator = RNGHandler.get_random_number_generator()

    def add(
        self,
        data_dictionary: Dict[str, np.ndarray],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        episode_ended = metadata.get(constants.METADATA_EPISODE_ENDED, None)
        if episode_ended is None:
            raise ValueError(
                "Episode ended metadata is required to add data to episodic replay buffer."
            )

        self._add_episodic_data(
            data_dictionary=data_dictionary, episode_ended=episode_ended
        )

        added_end_position = self.replay_buffer.get_next_index()

        # keep track of episode borders, needed for loading
        if added_end_position > self.last_added_end_position:
            self.buffer_episode_ended[
                self.last_added_end_position : added_end_position
            ] = episode_ended
        else:
            split = self._max_size - self.last_added_end_position
            self.buffer_episode_ended[self.last_added_end_position :] = episode_ended[
                :split
            ]
            self.buffer_episode_ended[:added_end_position] = episode_ended[split:]

        self.last_added_end_position = added_end_position

    def sample(self, batch_size: int):

        stored_size = self.size
        if stored_size < batch_size:
            # with replacement
            sample_ids = self.random_numbers_generator.choice(
                stored_size, size=batch_size, replace=True
            )
        else:
            # without replacement
            sample_ids = self.random_numbers_generator.choice(
                stored_size, size=batch_size, replace=False
            )

        return self.encode_sample(sample_ids)

    def encode_sample(self, sample_ids: List[int]) -> Dict[str, np.ndarray]:
        return self.replay_buffer._encode_sample(sample_ids)

    def save_checkpoint(self, context: SaveContext):
        if self.size == 0:
            # replay buffer is empty
            context.file_handler.save_to_file(
                to_save=self.FILE_EMPTY_SIGNAL,
                save_path=os.path.join(
                    context.save_path, context.prefix + self.FILE_REPLAY_BUFFER_HANDLING
                ),
            )
            return
        # Save replay buffer info

        replay_memory = self._calculate_memory_size()
        number_of_splits = math.ceil(replay_memory / self.REPLAY_SPLIT_SIZE)

        replay_buffer_size = self.replay_buffer.get_stored_size()
        split_size = math.ceil(replay_buffer_size / number_of_splits)
        rb_full = self._max_size <= self.replay_buffer.get_stored_size()

        handling_stats = (
            self.last_added_end_position,
            rb_full,
            self.buffer_episode_ended,
            number_of_splits,
            split_size,
        )
        context.file_handler.save_to_file(
            to_save=handling_stats,
            save_path=os.path.join(
                context.save_path,
                context.prefix + self.FILE_REPLAY_BUFFER_HANDLING,
            ),
        )

        # If replay buffer is to large to be saved in one go.
        if number_of_splits > 1:
            self._save_split_replay_buffer(
                file_handler=context.file_handler,
                save_path=context.save_path,
                prefix=context.prefix,
                number_of_splits=number_of_splits,
                split_size=split_size,
            )
        else:
            context.file_handler.save_to_file(
                to_save=self.replay_buffer.get_all_transitions(),
                save_path=os.path.join(
                    context.save_path, context.prefix + self.FILE_REPLAY_BUFFER
                ),
                enable_cache=False,
            )

    def load_checkpoint(self, context: SaveContext):

        loaded_handling_data = context.file_handler.load_from_file(
            load_path=os.path.join(
                context.save_path, context.prefix + self.FILE_REPLAY_BUFFER_HANDLING
            )
        )
        if loaded_handling_data == self.FILE_EMPTY_SIGNAL:
            # replay buffer is empty
            self.replay_buffer.clear()
            self.last_added_end_position = 0
            return
        # Load replay buffer info

        (
            self.last_added_end_position,
            rb_full,
            self.buffer_episode_ended,
            number_of_splits,
            split_size,
        ) = loaded_handling_data

        self.replay_buffer.clear()
        # add dummy data to set pointer to correct position
        if rb_full and not self.last_added_end_position == 0:
            self._add_dummy_data()

        # If replay buffer is to large to be loaded in one chunk
        if number_of_splits > 1:
            self._load_split_replay_buffer(
                file_handler=context.file_handler,
                load_path=context.save_path,
                prefix=context.prefix,
                number_of_splits=number_of_splits,
                split_size=split_size,
                rb_full=rb_full,
                environment_saved=context.metadata.get(
                    constants.SAVE_ARGUMENT_SAVE_ENVIRONMENT_STATE, False
                ),
            )
        else:
            self._load_cpprb(
                file_handler=context.file_handler,
                load_path=context.save_path,
                prefix=context.prefix,
                rb_full=rb_full,
                replay_buffer_info=self.replay_buffer_info,
                environment_saved=context.metadata.get(
                    constants.SAVE_ARGUMENT_SAVE_ENVIRONMENT_STATE, False
                ),
            )

    @property
    def size(self):
        return self.replay_buffer.get_stored_size()

    @property
    def max_size(self):
        return self._max_size

    @property
    def data_info(self):
        return self.replay_buffer_info

    @property
    def pointer_position(self):
        return self.last_added_end_position

    def _save_split_replay_buffer(
        self, file_handler, save_path, prefix, number_of_splits, split_size
    ):
        replay_buffer_size = self.replay_buffer.get_stored_size()
        for split in range(number_of_splits):
            split_start = split_size * split
            split_end = min(split_size * (split + 1), replay_buffer_size)

            split_ids = list(range(split_start, split_end))
            split_transition = self.replay_buffer._encode_sample(split_ids)
            file_path = os.path.join(
                save_path,
                prefix + self.FILE_REPLAY_BUFFER + "_split_" + str(split),
            )
            file_handler.save_to_file(
                to_save=split_transition, save_path=file_path, enable_cache=False
            )

            # Make sure the split is deleted from memory before loading the next one
            del split_transition
            gc.collect()

    def _add_non_episodic_data(self, data_dictionary: Dict[str, np.ndarray]) -> None:
        self.replay_buffer.add(**data_dictionary)

    def _add_episodic_data(
        self, data_dictionary: Dict[str, np.ndarray], episode_ended: np.ndarray
    ) -> None:
        batch_size = len(data_dictionary[self.some_data_key])

        episode_start_ids = np.where(episode_ended)[0] + 1
        transition_data_array_list = list(data_dictionary.values())

        # Create dictionary per episode
        episodic_data_list = map(
            lambda transition_data_array: np.split(
                transition_data_array, episode_start_ids, axis=0
            ),
            transition_data_array_list,
        )
        episodic_data_dictionaries = list(
            map(
                lambda episodic_data: dict(zip(data_dictionary.keys(), episodic_data)),
                zip(*episodic_data_list),
            )
        )

        # Add data of ending episodes
        for i in range(len(episode_start_ids)):
            self.replay_buffer.add(**episodic_data_dictionaries[i])
            self.replay_buffer.on_episode_end()

        # Add remaining data without ending episode
        if episode_start_ids.size < 1 or episode_start_ids[-1] < batch_size:
            self.replay_buffer.add(**episodic_data_dictionaries[-1])

    def _add_dummy_data(self):
        dummy_transitions = {}
        for key in self.replay_buffer_info.keys():
            dummy_transitions[key] = np.zeros(
                (self.last_added_end_position, *self.replay_buffer_info[key]["shape"])
            )

        self.replay_buffer.add(**dummy_transitions)
        self.replay_buffer.on_episode_end()

        del dummy_transitions
        gc.collect()

    def _load_split_replay_buffer(
        self,
        file_handler,
        load_path,
        prefix,
        number_of_splits,
        split_size,
        rb_full,
        environment_saved=False,
    ):
        # We dont need to shift the data if the replay buffer is not full,
        # or when it is full but the pointer is at the beginning of the replay buffer
        shift_needed = rb_full and self.last_added_end_position != 0

        if shift_needed:
            shift_split = self.last_added_end_position // split_size
            # shift split ids according to pointer
            splits = deque(range(number_of_splits))
            splits.rotate(-shift_split)
            splits = list(splits)
            shifted_buffer_episode_ended = np.concatenate(
                [
                    self.buffer_episode_ended[self.last_added_end_position :],
                    self.buffer_episode_ended[: self.last_added_end_position],
                ]
            )
            added_so_far = 0

            # load first Split in which the data needs to be shifted according to pointer
            split_transitions = file_handler.load_from_file(
                load_path=os.path.join(
                    load_path,
                    prefix + self.FILE_REPLAY_BUFFER + "_split_" + str(splits[0]),
                ),
                enable_cache=False,
            )
            shift_id = self.last_added_end_position - split_size * splits[0]
            data = {
                key: split_transitions[key][shift_id:]
                for key in self.replay_buffer_info.keys()
            }
            # The data that is right in front of the pointer needs to be added last
            to_add_last_data = {
                key: split_transitions[key][:shift_id]
                for key in self.replay_buffer_info.keys()
            }
            first_split_shift_length = len(data[self.some_data_key])
            # If pointer is right at the beginning of the split, no data needs to be shifted
            if first_split_shift_length > 0:
                self._add_episodic_data(
                    data_dictionary=data,
                    episode_ended=shifted_buffer_episode_ended[
                        :first_split_shift_length
                    ],
                )
                added_so_far += first_split_shift_length

            del data
            gc.collect()

            remaining_splits = splits[1:]
        else:
            # If no shift is needed, we can load the data as is
            added_so_far = 0
            remaining_splits = range(number_of_splits)
            shifted_buffer_episode_ended = self.buffer_episode_ended

        # Tell the replay buffer that due to the loading the episode was truncated, if the environment is not saved
        if not environment_saved:
            shifted_buffer_episode_ended[-1] = True

        # load all other splits
        for split in remaining_splits:
            data = {}
            split_transitions = file_handler.load_from_file(
                load_path=os.path.join(
                    load_path,
                    prefix + self.FILE_REPLAY_BUFFER + "_split_" + str(split),
                ),
                enable_cache=False,
            )
            split_size = len(split_transitions[self.some_data_key])
            self._add_episodic_data(
                data_dictionary=split_transitions,
                episode_ended=shifted_buffer_episode_ended[
                    added_so_far : added_so_far + split_size
                ],
            )
            added_so_far += split_size

            del data
            gc.collect()

        # add last bit of first split
        if shift_needed:
            if len(to_add_last_data[self.some_data_key]) > 0:
                self._add_episodic_data(
                    data_dictionary=to_add_last_data,
                    episode_ended=shifted_buffer_episode_ended[added_so_far:],
                )

            del to_add_last_data
            gc.collect()

    def _load_cpprb(
        self,
        file_handler,
        load_path,
        prefix,
        rb_full,
        replay_buffer_info,
        environment_saved=False,
    ):
        all_transitions = file_handler.load_from_file(
            load_path=os.path.join(load_path, prefix + self.FILE_REPLAY_BUFFER)
        )
        data_dict = {}
        if rb_full:
            for key in replay_buffer_info.keys():
                # shift data according to pointer
                data_dict[key] = np.vstack(
                    [
                        all_transitions[key][self.last_added_end_position :],
                        all_transitions[key][: self.last_added_end_position],
                    ]
                )
                shifted_episode_ended_bools = np.concatenate(
                    [
                        self.buffer_episode_ended[self.last_added_end_position :],
                        self.buffer_episode_ended[: self.last_added_end_position],
                    ]
                )
            # Tell the replay buffer that due to the loading the last episode was truncated if the environment is not saved
            if not environment_saved:
                shifted_episode_ended_bools[-1] = True
        else:
            data_dict = all_transitions
            shifted_episode_ended_bools = self.buffer_episode_ended
            # Tell the replay buffer that due to the loading the last episode was truncated if the environment is not saved
            num_entries = len(data_dict[self.some_data_key])
            if not environment_saved:
                shifted_episode_ended_bools[num_entries - 1] = True

        self._add_episodic_data(
            data_dictionary=data_dict, episode_ended=shifted_episode_ended_bools
        )

        del data_dict
        gc.collect()

    def _calculate_memory_size(self) -> int:

        bytes_per_transition = 0
        transition = self.replay_buffer.sample(1)
        for key in self.replay_buffer_info.keys():

            bytes_per_transition += sys.getsizeof(deepcopy(transition[key][0]))

        return bytes_per_transition * self.replay_buffer.get_stored_size()

    def _create_initialization_replay_buffer_info(
        self, replay_buffer_info: Dict, additional_arguments: Dict[str, Any]
    ) -> Dict:
        reduced_replay_buffer_info = deepcopy(replay_buffer_info)
        # remove next_of from replay_buffer_info
        if "next_of" in additional_arguments.keys():
            if isinstance(additional_arguments["next_of"], list):
                for key in additional_arguments["next_of"]:
                    if key not in replay_buffer_info.keys():
                        raise KeyError("next_of key {key} is not in replay_buffer_info")
                    del reduced_replay_buffer_info[f"next_{key}"]
            else:
                if additional_arguments["next_of"] not in replay_buffer_info.keys():
                    raise KeyError(
                        f"next_of key {additional_arguments['next_of']} is not in replay_buffer_info"
                    )
                del reduced_replay_buffer_info[
                    f"next_{additional_arguments['next_of']}"
                ]

        return reduced_replay_buffer_info



from typing import Dict, Any, Tuple

from gymnasium.spaces import Space, Discrete

import athlete
from athlete.data_collection.collector import DataCollector
from athlete.data_collection.transition import GymnasiumTransitionDataCollector
from athlete.update.update_rule import UpdateRule
from athlete.algorithms.q_learning.update import QLearningUpdate
from athlete.policy.policy_builder import PolicyBuilder
from athlete.algorithms.q_learning.policy import QLearningPolicyBuilder
from athlete.data_collection.provider import UpdateDataProvider
from athlete import constants

# Q-Learning

ARGUMENT_DISCOUNT = "discount"
ARGUMENT_LEARNING_RATE = "learning_rate"
ARGUMENT_START_EPSILON = "start_epsilon"
ARGUMENT_END_EPSILON = "end_epsilon"
ARGUMENT_EPSILON_DECAY_STEPS = "epsilon_decay_steps"

DEFAULT_CONFIGURATION = {
    ARGUMENT_DISCOUNT: 0.9,
    ARGUMENT_LEARNING_RATE: 0.1,
    ARGUMENT_START_EPSILON: 1.0,
    ARGUMENT_END_EPSILON: 0.1,
    constants.GENERAL_ARGUMENT_WARMUP_STEPS: 0,
    ARGUMENT_EPSILON_DECAY_STEPS: 100,
}


def make_q_learning_components(
    observation_space: Space, action_space: Space, configuration: Dict[str, Any]
) -> Tuple[DataCollector, UpdateRule, PolicyBuilder]:
    """Creates the components for a Q-learning agent.

    Args:
        observation_space: Environment observation space
        action_space: Environment action space
        configuration: Algorithm configuration dictionary

    Returns:
        Tuple containing data collector, update rule and policy builder

    Raises:
        ValueError: If observation_space or action_space is not Discrete
    """

    if not isinstance(observation_space, Discrete):
        raise ValueError(
            f"This Q-Learning implementation only supports {Discrete.__name__} observation spaces, but got {type(observation_space)}"
        )
    if not isinstance(action_space, Discrete):
        raise ValueError(
            f"This Q-Learning implementation only supports {Discrete.__name__} action spaces, but got {type(action_space)}"
        )

    update_data_provider = UpdateDataProvider()

    # DATA COLLECTOR
    data_collector = GymnasiumTransitionDataCollector(
        update_data_provider=update_data_provider,
    )

    # UPDATE RULE

    update_rule = QLearningUpdate(
        observation_space=observation_space,
        action_space=action_space,
        update_data_provider=update_data_provider,
        discount=configuration[ARGUMENT_DISCOUNT],
        learning_rate=configuration[ARGUMENT_LEARNING_RATE],
    )

    # POLICY BUILDER

    policy_builder = QLearningPolicyBuilder(
        q_table=update_rule.q_table,
        action_space=action_space,
        start_epsilon=configuration[ARGUMENT_START_EPSILON],
        end_epsilon=configuration[ARGUMENT_END_EPSILON],
        epsilon_decay_steps=configuration[ARGUMENT_EPSILON_DECAY_STEPS],
    )

    return data_collector, update_rule, policy_builder


athlete.register(
    id="q_learning",
    component_factory=make_q_learning_components,
    default_configuration=DEFAULT_CONFIGURATION,
)



from typing import Callable, Dict, Any, Type, Optional, Tuple
import copy

import torch
from torch import nn
from torch.optim import Optimizer
from gymnasium.spaces import Space

from athlete import constants
from athlete.update.update_rule import UpdateRule, UpdatableComponent
from athlete.algorithms.ddpg.updatable_components import (
    DDPGCriticUpdate,
    DDPGActorUpdate,
)
from athlete.update.common import TargetNetUpdate
from athlete.saving.saveable_component import CompositeSaveableComponent
from athlete.data_collection.provider import UpdateDataProvider
from athlete.update.buffer import EpisodicCPPReplayBuffer
from athlete.update.common import TargetNetUpdate, ReplayBufferUpdate
from athlete.update.buffer_wrapper import (
    PostBufferPreprocessingWrapper,
)
from athlete.function import extract_data_from_batch, create_transition_data_info


class DDPGUpdate(UpdateRule, CompositeSaveableComponent):
    """Update rule for the DDPG algorithm.
    This class manages all updatable components of the DDPG algorithm.
    It also handles the saving off all stateful objects of the DDPG algorithm.
    """

    def __init__(
        self,
        observation_space: Space,
        action_space: Space,
        update_data_input: UpdateDataProvider,
        critic: nn.Module,
        actor: nn.Module,
        discount: float,
        critic_optimizer_class: Type[Optimizer],
        actor_optimizer_class: Type[Optimizer],
        critic_optimizer_arguments: Dict[str, Any],
        actor_optimizer_arguments: Dict[str, Any],
        critic_criteria: torch.nn.modules.loss._Loss,
        replay_buffer_capacity: int,
        replay_buffer_mini_batch_size: int,
        critic_update_frequency: int,
        critic_number_of_updates: int,
        actor_update_frequency: int,
        actor_number_of_updates: int,
        multiply_number_of_updates_by_environment_steps: bool,
        target_critic_update_frequency: int,
        target_actor_update_frequency: int,
        target_critic_tau: Optional[float],
        target_actor_tau: Optional[float],
        critic_gradient_max_norm: Optional[float],
        actor_gradient_max_norm: Optional[float],
        device: str,
        additional_replay_buffer_arguments: Dict[str, Any],
        post_replay_buffer_data_preprocessing: Optional[Dict[str, Callable]],
    ) -> None:
        """
        Initializes the DDPG update rule.

        Args:
            observation_space (Space): The observation space of the environment.
            action_space (Space): The action space of the environment.
            update_data_input (UpdateDataProvider): The data provider for the update rule. This is used
                to communicate with the data collector.
            critic (nn.Module): The critic network.
            actor (nn.Module): The actor network.
            discount (float): The discount factor for the critic.
            critic_optimizer_class (Type[Optimizer]): The class of the critic optimizer, a subclass of torch.optim.Optimizer.
            actor_optimizer_class (Type[Optimizer]): The class of the actor optimizer, a subclass of torch.optim.Optimizer.
            critic_optimizer_arguments (Dict[str, Any]): The arguments for initializing the critic optimizer, not including the parameters critic parameters.
            actor_optimizer_arguments (Dict[str, Any]): The arguments for initializing the actor optimizer, not including the parameters actor parameters.
            critic_criteria (torch.nn.modules.loss._Loss): The loss function used for calculating the loss between the predicted and target values for the critic.
            replay_buffer_capacity (int): The maximum capacity of the replay buffer.
            replay_buffer_mini_batch_size (int): The mini-batch size for sampling from the replay buffer.
            critic_update_frequency (int): The frequency of the critic update according to the number of environment steps. If -1 the critic is updated at the end of each episode.
            critic_number_of_updates (int): The number of updates for the critic on each update step.
            actor_update_frequency (int): The frequency of the actor update according to the number of environment steps. If -1 the actor is updated at the end of each episode.
            actor_number_of_updates (int): The number of updates for the actor on each update step.
            multiply_number_of_updates_by_environment_steps (bool): Whether the number of updates for actor and critic is multiplied by the environment steps since the last update.
            target_critic_update_frequency (int): The frequency of the target critic update according to the number of environment steps.
            target_actor_update_frequency (int): The frequency of the target actor update according to the number of environment steps.
            target_critic_tau (Optional[float]) : The tau parameter for the target critic update. If none, a hard update is performed.
            target_actor_tau (Optional[float]): The tau parameter for the target actor update. If none, a hard update is performed.
            critic_gradient_max_norm (Optional[float]): The maximum norm for the critic gradient clipping. If None, no gradient clipping is performed.
            actor_gradient_max_norm (Optional[float]): The maximum norm for the actor gradient clipping. If None, no gradient clipping is performed.
            device (str): The device to use for the update rule. This is used to move the networks and data to the correct device.
            additional_replay_buffer_arguments (Dict[str, Any]): Additional arguments for the replay buffer initialization. Might be used for specific memory compression.
            post_replay_buffer_data_preprocessing (Optional[Dict[str, Callable]]): A dictionary of functions to preprocess the data sampled from the replay buffer. The keys are the field names of the data and the values are the functions to apply.
        """

        super().__init__()

        UpdateRule.__init__(self)
        CompositeSaveableComponent.__init__(self)

        # Updatable components
        # General stateful components

        self.critic = critic.to(device)
        self.actor = actor.to(device)

        self.critic_optimizer = critic_optimizer_class(
            params=self.critic.parameters(), **critic_optimizer_arguments
        )
        self.actor_optimizer = actor_optimizer_class(
            params=self.actor.parameters(), **actor_optimizer_arguments
        )

        self.register_saveable_component("critic", self.critic)
        self.register_saveable_component("actor", self.actor)
        self.register_saveable_component("critic_optimizer", self.critic_optimizer)
        self.register_saveable_component("actor_optimizer", self.actor_optimizer)

        self.target_critic = copy.deepcopy(self.critic).eval()
        self.target_actor = copy.deepcopy(self.actor).eval()

        self.register_saveable_component("target_critic", self.target_critic)
        self.register_saveable_component("target_actor", self.target_actor)

        # Replay Buffer Update

        update_data_info = create_transition_data_info(
            observation_space=observation_space,
            action_space=action_space,
        )

        additional_arguments = {
            "next_of": constants.DATA_OBSERVATIONS,
        }
        additional_arguments.update(additional_replay_buffer_arguments)

        self.replay_buffer = EpisodicCPPReplayBuffer(
            capacity=replay_buffer_capacity,
            replay_buffer_info=update_data_info,
            additional_arguments=additional_arguments,
        )

        self.register_saveable_component("replay_buffer", self.replay_buffer)

        self.replay_buffer_update = ReplayBufferUpdate(
            update_data_provider=update_data_input,
            replay_buffer=self.replay_buffer,
        )

        if post_replay_buffer_data_preprocessing is not None:
            sample_replay_buffer = PostBufferPreprocessingWrapper(
                replay_buffer=self.replay_buffer,
                post_replay_buffer_data_preprocessing=post_replay_buffer_data_preprocessing,
            )
        else:
            sample_replay_buffer = self.replay_buffer

        # Critic Update

        critic_data_keys = list(update_data_info.keys())
        critic_data_sampler_function = lambda: extract_data_from_batch(
            sample_replay_buffer.sample(replay_buffer_mini_batch_size),
            keys=critic_data_keys,
            device=device,
        )

        self.critic_update = DDPGCriticUpdate(
            data_sampler=critic_data_sampler_function,
            critic=self.critic,
            target_critic=self.target_critic,
            critic_optimizer=self.critic_optimizer,
            target_actor=self.target_actor,
            discount=discount,
            changes_policy=False,
            update_frequency=critic_update_frequency,
            number_of_updates=critic_number_of_updates,
            multiply_number_of_updates_by_environment_steps=(
                multiply_number_of_updates_by_environment_steps
            ),
            criteria=critic_criteria,
            log_tag=DDPGCriticUpdate.CRITIC_LOSS_LOG_TAG,
            gradient_max_norm=critic_gradient_max_norm,
        )

        # Actor Update
        actor_data_keys = [constants.DATA_OBSERVATIONS]
        actor_data_sampler_function = lambda: extract_data_from_batch(
            sample_replay_buffer.sample(replay_buffer_mini_batch_size),
            keys=actor_data_keys,
            device=device,
        )

        self.actor_update = DDPGActorUpdate(
            data_sampler=actor_data_sampler_function,
            actor=self.actor,
            actor_optimizer=self.actor_optimizer,
            critic=self.critic,
            changes_policy=True,
            update_frequency=actor_update_frequency,
            number_of_updates=actor_number_of_updates,
            multiply_number_of_updates_by_environment_steps=(
                multiply_number_of_updates_by_environment_steps
            ),
            log_tag=DDPGActorUpdate.ACTOR_LOSS_LOG_TAG,
            gradient_max_norm=actor_gradient_max_norm,
        )

        # Target Critic Update
        self.target_critic_update = TargetNetUpdate(
            source_net=self.critic,
            target_net=self.target_critic,
            tau=target_critic_tau,
            changes_policy=False,
            update_frequency=target_critic_update_frequency,
        )

        # Target Actor Update
        self.target_actor_update = TargetNetUpdate(
            source_net=self.actor,
            target_net=self.target_actor,
            tau=target_actor_tau,
            changes_policy=False,
            update_frequency=target_actor_update_frequency,
        )

    @property
    def updatable_components(self) -> Tuple[UpdatableComponent]:
        """Returns all updatable components of the DDPG algorithm in the order they are
        updated.

        Returns:
            Tuple[UpdatableComponent]: A tuple of updatable components.
                1. Replay Buffer Update
                2. Critic Update
                3. Actor Update
                4. Target Critic Update
                5. Target Actor Update
        """
        return (
            self.replay_buffer_update,
            self.critic_update,
            self.actor_update,
            self.target_critic_update,
            self.target_actor_update,
        )



from operator import itemgetter
from functools import reduce
from typing import Union, List, Callable, Any, Tuple, Dict

import numpy as np
import torch
from gymnasium.spaces import Space

from athlete import constants

# Specific dtype mapping for numpy to torch conversion
# This assumes that our torch implementation uses float32 for most operations
# Mapping most ints to int64 because torch can only perform indexing with int64
# Mapping int8 to int8 because we assume this is only used if you want to save memory
# By default bools are to integers, this mapping maps booleans to booleans
DTYPE_MAP = {
    "float16": torch.float32,
    "float32": torch.float32,
    "float64": torch.float32,
    "int8": torch.int64,
    "int16": torch.int64,
    "int32": torch.int64,
    "int64": torch.int64,
    "uint8": torch.uint8,
    "bool": torch.bool,
}


def numpy_to_tensor(np_array: np.ndarray, device: str = "cpu") -> torch.Tensor:
    """Transforms a numpy array to a torch tensor following a specific dtype mapping.

    Args:
        np_array (np.ndarray): Numpy array to be transformed.
        device (str, optional): Device to which the tensor should be moved. Defaults to "cpu".

    Returns:
        torch.Tensor: Transformed torch tensor.
    """
    dtype = DTYPE_MAP.get(np_array.dtype.name, torch.float32)
    return (
        torch.from_numpy(np_array).to(device=device, dtype=dtype).requires_grad_(False)
    )


def tensor_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    """Safely transforms a torch tensor to a numpy array. Uses default dtype mapping from numpy.

    Args:
        tensor (torch.Tensor): Torch tensor to be transformed.

    Returns:
        np.ndarray: Transformed numpy array.
    """
    return tensor.detach().cpu().numpy()


def gymnasium_value_to_batched_numpy_array(value: Union[int, np.ndarray]) -> np.ndarray:
    """Takes a value that might be a numpy array or an int and and returns a numpy array with a batch dimension.
    Also copies the data to ensure that the original data is not modified by following operations.

    Args:
        value (Union[int, np.ndarray]): Value to be transformed. Can be a numpy array or an int.

    Returns:
        np.ndarray: Transformed numpy array with a batch dimension.
    """
    if isinstance(value, np.ndarray):
        value = value.copy()
    else:
        value = np.array([value])

    value = np.expand_dims(value, axis=0)

    return value


def single_safe_itemgetter(keys: List[str]) -> Callable[[Any], Tuple[Any]]:
    """
    An Itemgetter that always returns a tuple even if only one key is provided.

    Args:
        keys (list): List of keys.

    Returns:
        callable: A Itemgetter function that always returns a tuple.
    """
    if len(keys) > 1:
        return itemgetter(*keys)
    else:
        return lambda _dict: (itemgetter(*keys)(_dict),)


def chain_functions(
    function_list: List[Callable], input_value: Any
) -> Callable[[Any], Any]:
    """Returns a function that takes an input value and applies a list of functions to it in order.

    Args:
        function_list (_type_): List of functions to be applied to the input value.
        input_value (_type_): Input value to be passed through the functions in succession.

    Returns:
        Callable[[Any], Any]: A function that takes an input value and applies the list of functions to it in order.
    """
    return (
        reduce(
            lambda intermediate_value, function: function(intermediate_value),
            function_list,
            input_value,
        )
        if function_list
        else input_value
    )


def extract_data_from_batch(
    data_batch: Dict[str, np.ndarray], keys: List[str], device: str
) -> Dict[str, torch.Tensor]:
    """Extracts data from a batch and converts it to torch tensors.

    Args:
        data_batch (Dict[str, np.ndarray]): Batch of data to be extracted.
        keys (List[str]): List of keys to extract from the batch.
        device (str): Device to which the tensors should be moved.

    Returns:
        Dict[str, torch.Tensor]: Extracted data as torch tensors.
    """
    return dict(
        zip(
            keys,
            map(
                lambda data: numpy_to_tensor(data, device=device),
                single_safe_itemgetter(keys)(data_batch),
            ),
        )
    )


def create_transition_data_info(
    observation_space: Space, action_space: Space
) -> Dict[str, Dict[str, Any]]:
    """Creates a dictionary with data information of a transition used for the replay buffer according to the observation and action space.

    Args:
        observation_space (Space): Observation space of the data.
        action_space (Space): Action space of the data.

    Returns:
        Dict[str, Dict[str, Any]]: Dictionary with data information of a transition.
    """

    observation_info = {
        "shape": observation_space.shape,
        "dtype": str(observation_space.dtype),
    }
    if observation_info["shape"] == ():
        observation_info["shape"] = (1,)

    action_info = {"shape": action_space.shape, "dtype": str(action_space.dtype)}
    if action_info["shape"] == ():
        action_info["shape"] = (1,)

    return {
        constants.DATA_REWARDS: {"shape": (1,), "dtype": np.float32},
        constants.DATA_OBSERVATIONS: observation_info,
        constants.DATA_NEXT_OBSERVATIONS: observation_info,
        constants.DATA_ACTIONS: action_info,
        constants.DATA_TERMINATEDS: {"shape": (1,), "dtype": np.bool_},
    }



from abc import ABC, abstractmethod


class DataCollector(ABC):
    """An Interface for data collection.
    This is on of the three main components of an algorithm.
    It takes in the data given to the agent and does with it what ever is needed such that the
    update rule can use the data.
    Note that there is no passing of values between the data collector and the update rule.
    All communication is done via injected dependencies.
    """

    @abstractmethod
    def collect(self, *args, **kwargs) -> bool:
        """Collecting data from the environment during a regular step.

        Returns:
            bool: Whether a new datapoint has been created during this collection for not.
        """
        ...

    @abstractmethod
    def collect_reset(self, *args, **kwargs) -> bool:
        """Collecting data from the environment on a reset call of the environment.
        This does not mean that the collector is reset.

        Returns:
            bool: Whether a new datapoint has been created during this collection for not.
        """
        ...



import torch
from torch import nn
from typing import Callable, Dict, Any, Tuple, Optional
import copy

from gymnasium.spaces import Box, Discrete

from athlete import constants
from athlete.update.update_rule import UpdateRule, UpdatableComponent
from athlete.algorithms.dqn.updatable_components import DQNValueUpdate
from athlete.update.common import TargetNetUpdate, ReplayBufferUpdate
from athlete.data_collection.provider import UpdateDataProvider
from athlete.update.buffer import EpisodicCPPReplayBuffer
from athlete.update.buffer_wrapper import (
    PostBufferPreprocessingWrapper,
)
from athlete.function import extract_data_from_batch, create_transition_data_info
from athlete.saving.saveable_component import CompositeSaveableComponent


class DQNUpdate(UpdateRule, CompositeSaveableComponent):
    """The Update rule for DQN.
    Manages all updatable components and the saving and loading of stateful objects.
    """

    def __init__(
        self,
        observation_space: Box,
        action_space: Discrete,
        update_data_input: UpdateDataProvider,
        q_value_function: nn.Module,
        discount: float,
        optimizer_class: torch.optim.Optimizer,
        optimizer_arguments: Dict[str, Any],
        replay_buffer_capacity: int,
        replay_buffer_mini_batch_size: int,
        value_net_update_frequency: int,
        value_net_number_of_updates: int,
        multiply_number_of_updates_by_environment_steps: bool,
        target_net_update_frequency: int,
        target_net_tau: Optional[float],
        enable_double_q_learning: bool,
        criteria: torch.nn.modules.loss._Loss,
        gradient_max_norm: Optional[float],
        device: str,
        additional_replay_buffer_arguments: Dict[str, Any],
        post_replay_buffer_data_preprocessing: Optional[Dict[str, Callable]],
    ) -> None:
        """Initializes the DQN update rule.

        Args:
            observation_space (Box): Observation space of the environment.
            action_space (Discrete): Action space of the environment.
            update_data_input (UpdateDataProvider): Data provider used to communicate with the data collector.
            q_value_function (nn.Module): The Q-value function network.
            discount (float): The discount factor for the Q-value function.
            optimizer_class (torch.optim.Optimizer): The optimizer class to use for the Q-value function.
            optimizer_arguments (Dict[str, Any]): The initialization arguments for the optimizer with the Q-value function parameters.
            replay_buffer_capacity (int): The maximum capacity of the replay buffer.
            replay_buffer_mini_batch_size (int): The mini-batch size for sampling from the replay buffer.
            value_net_update_frequency (int): The frequency of updates for the Q-value function according to the number of environment steps.
                If -1, the update is performed at the end of each episode.
            value_net_number_of_updates (int): The number of updates to perform at each update step.
            multiply_number_of_updates_by_environment_steps (bool): Whether to multiply the number of updates by the number of environment steps since the last update.
            target_net_update_frequency (int): The frequency of updates for the target network.
            target_net_tau (Optional[float]): The soft update factor for the target network if none, uses hard updates instead.
            enable_double_q_learning (bool): Whether to use double Q-learning for the target calculation.
            criteria (torch.nn.modules.loss._Loss): The loss function used to calculate the loss based on prediction and target.
            gradient_max_norm (Optional[float]): The maximum norm for gradient clipping. If None, no clipping is performed.
            device (str): The device on which the updates are performed (e.g., "cpu" or "cuda").
            additional_replay_buffer_arguments (Dict[str, Any]): Additional arguments for initializing the replay buffer.
                This might be used for specific memory compression techniques.
            post_replay_buffer_data_preprocessing (Optional[Dict[str, Callable]]):
                A dictionary of functions to preprocess the data before passing it to the updates.
                The keys are the names of the data fields, and the values are the functions themselves.
        """

        UpdateRule.__init__(self)
        CompositeSaveableComponent.__init__(self)

        # General stateful components

        self.q_value_function = q_value_function.to(device)

        self.optimizer = optimizer_class(
            params=q_value_function.parameters(), **optimizer_arguments
        )
        self.register_saveable_component("value_function", self.q_value_function)
        self.register_saveable_component("optimizer", self.optimizer)

        # Replay Buffer Update

        additional_arguments = {
            "next_of": constants.DATA_OBSERVATIONS,
        }
        additional_arguments.update(additional_replay_buffer_arguments)

        update_data_info = create_transition_data_info(
            observation_space=observation_space,
            action_space=action_space,
        )

        self.replay_buffer = EpisodicCPPReplayBuffer(
            capacity=replay_buffer_capacity,
            replay_buffer_info=update_data_info,
            additional_arguments=additional_arguments,
        )

        self.register_saveable_component("replay_buffer", self.replay_buffer)

        self.replay_buffer_update = ReplayBufferUpdate(
            update_data_provider=update_data_input,
            replay_buffer=self.replay_buffer,
        )

        if post_replay_buffer_data_preprocessing is not None:
            sample_replay_buffer = PostBufferPreprocessingWrapper(
                replay_buffer=self.replay_buffer,
                post_replay_buffer_data_preprocessing=post_replay_buffer_data_preprocessing,
            )
        else:
            sample_replay_buffer = self.replay_buffer

        # Value function Update
        self.target_net = copy.deepcopy(q_value_function).eval()
        self.register_saveable_component("target_net", self.target_net)

        extract_keys = list(update_data_info.keys())
        data_sampler_function = lambda: extract_data_from_batch(
            sample_replay_buffer.sample(replay_buffer_mini_batch_size),
            keys=extract_keys,
            device=device,
        )

        self.value_function_update = DQNValueUpdate(
            q_value_function=q_value_function,
            target_net=self.target_net,
            optimizer=self.optimizer,
            data_sampler=data_sampler_function,
            cross_validation=enable_double_q_learning,
            discount=discount,
            criteria=criteria,
            log_tag=DQNValueUpdate.LOG_TAG_LOSS,
            number_of_updates=value_net_number_of_updates,
            multiply_number_of_updates_by_environment_steps=multiply_number_of_updates_by_environment_steps,
            gradient_max_norm=gradient_max_norm,
            update_frequency=value_net_update_frequency,
            changes_policy=True,
        )

        # This is required because dqn value update might needs to know during which step it was last updated to calculated number of updates.
        self.register_saveable_component("value_update", self.value_function_update)

        # Target Net Update

        self.target_net_update = TargetNetUpdate(
            source_net=q_value_function,
            target_net=self.target_net,
            tau=target_net_tau,
            update_frequency=target_net_update_frequency,
        )

    @property
    def updatable_components(self) -> Tuple[UpdatableComponent]:
        """Returns all updatable components of the update rule in the order they should be updated.

        Returns:
            Tuple[UpdatableComponent]: A tuple of all updatable components:
                1. Replay buffer update
                2. Value function update
                3. Target network update
        """
        return (
            self.replay_buffer_update,
            self.value_function_update,
            self.target_net_update,
        )
